/**
 * \file Address_Space.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date May 2005
 * \brief The code for the Address_Space class.
 *
 * Defines the methods within an Address_Space object, which itself
 * represents the sequence of adressable spaces (divided into pages)
 * associated with a given process, file, or shm segment.
 **/



// ===================================================================
// INCLUDES

#include "Address_Space.hh"
#include "Canonical_Page.hh"
// ===================================================================



// ===================================================================
void
Address_Space::insert_mapping (const virtual_page_ID_t virtual_page_ID,
			       Canonical_Page* canonical_page) {

  // Sanity check: We should be mapping a real canonical page.
  assert(canonical_page != NULL);

  // Sanity check: If this address space is a file, then the virtual
  // page number must be within the number of pages in the file.
  assert((type != AS_TYPE_FILE) || (virtual_page_ID < size));

  // There should not be an existing mapping.
  assert(map.find(virtual_page_ID) == map.end());

  // Mappings are not performed for uncacheable spaces.
  if (type == AS_TYPE_UNCACHED) {

    // Since we are not mapping it here, we may be able to delete the
    // provided canonical page.
    delete_if_unused_page(&canonical_page);

  } else {

    // Insert the new mapping.
    map[virtual_page_ID] = canonical_page;
    canonical_page->increment_mapping_count();

  }

} // Address_Space::insert_mapping
// ===================================================================



// ===================================================================
void
Address_Space::remove_mapping (const virtual_page_ID_t virtual_page_ID) {

  // Sanity check: If this address space is a file, then the virtual
  // page number must be within the number of pages in the file.
  assert((type != AS_TYPE_FILE) || (virtual_page_ID < size));

  // Grab an iterator to the mapping.  There is only work to do if the
  // mapping exists.
  virtual_to_canonical_map_t::iterator i = map.find(virtual_page_ID);
  if (i != map.end()) {

    // Remove the mapping, eliminating the canonical page if it has
    // become garbage.
    Canonical_Page* canonical_page = (*i).second;
    map.erase(i);
    canonical_page->decrement_mapping_count();
    delete_if_unused_page(&canonical_page);

  }

} // Address_Space::remove_mapping
// ===================================================================



// ===================================================================
Canonical_Page*
Address_Space::lookup_canonical_page (virtual_page_ID_t virtual_page_ID) {
	
  Canonical_Page* canonical_page = NULL;

  // If the space is uncacheable, simply return the uncachable page.
  // Otherwise, attempt to look up the given virtual page ID.
  if (type == AS_TYPE_UNCACHED) {

    canonical_page = &canonical_uncacheable_page;

  } else {

    // Attempt to lookup the given virtual page ID.
    virtual_to_canonical_map_t::iterator i = map.find(virtual_page_ID);
    if (i != map.end()) {
		
      // Found it!
      canonical_page = (*i).second;

    }

  }

  return canonical_page;
	
} // Address_Space::lookup_canonical_page
// ===================================================================



// ===================================================================
Canonical_Page*
Address_Space::safe_lookup_canonical_page
(const virtual_page_ID_t virtual_page_ID) {
	
  // Attempt to lookup the page.
  Canonical_Page* canonical_page =
    lookup_canonical_page(virtual_page_ID);
	
  // Was the lookup successful?
  if (canonical_page == NULL) {
		
    // No.  This is an invalid virtual page ID for this address space.
    // Emit the error and exit.
    fprintf(stderr, "Address_Space::safe_lookup_canonical_page(): ");
    fprintf(stderr, "Failed lookup\n");
    exit(1);
		
  }
	
  return canonical_page;
	
} // Address_Space::safe_lookup_canonical_page
// ===================================================================



// ===================================================================
void
Address_Space::expand_size (const file_offset_t size) {

  // We only track sizes for files, so update only if this is an
  // expansion for such a space.
  if ((type == AS_TYPE_FILE) && (size > this->size)) {

    this->size = size;

  }

} // Address_Space::expand_size
// ===================================================================



// ===================================================================
void
Address_Space::contract_size (const file_offset_t size) {

  // We only track sizes for files.
  if (type == AS_TYPE_FILE) {

    // Sanity check: A contraction should never be requested for a
    // size that is larger than the known one.
    assert(size <= this->size);
    this->size = size;

  }

} // Address_Space::contract_size
// ===================================================================



// ===================================================================
file_offset_t
Address_Space::get_size () const {

  // Sanity check: Sizes are only meaningful for file address spaces.
  assert(type == AS_TYPE_FILE);

  return size;

} // Address_Space::get_size
// ===================================================================



// ===================================================================
address_space_type_t
Address_Space::get_type () const {

  return type;

} // Address_Space::get_type
// ===================================================================



// ===================================================================
void
Address_Space::test_type
  (const address_space_type_t address_space_type) const {

  if (address_space_type != type) {
    fprintf(stderr, "Address_Space::test_type(): ");
    fprintf(stderr, "Mismatched types\n");
    exit(1);
  }
	
} // Address_Space::test_type
// ===================================================================



// ===================================================================
void
Address_Space::test_file_type (file_type_t file_type) const {
	
  // Based on the given file type, test the address space
  // accordingly.  Note that if the file type given is `unknown', then
  // no comparison is performed.
  switch (file_type) {
		
  case FILE_TYPE_REGULAR:
    test_type(AS_TYPE_FILE);
    break;
			
  case FILE_TYPE_PIPE:
  case FILE_TYPE_DEVICE:
  case FILE_TYPE_SOCKET:
    test_type(AS_TYPE_UNCACHED);
    break;

  case FILE_TYPE_UNKNOWN:
    break;
			
  default:
    fprintf(stderr, "Address_Space::test_file_type(): ");
    fprintf(stderr, "Unknown file type\n");
    exit(1);
			
  }

} // Address_Space::test_file_type
// ===================================================================



// ===================================================================
void
Address_Space::clear_all_mappings () {
	
  // For uncached spaces, there's nothing to clear.
  if (type == AS_TYPE_UNCACHED) {
		
    // Sanity check: There should be no mappings.
    assert(map.begin() == map.end());
    return;
		
  }
	
  // Traverse the map, unlinking virtual and canonical pages.
  virtual_to_canonical_map_t::iterator i = map.begin();
  while (i != map.end()) {
		
    Canonical_Page* canonical_page = (*i).second;
		
    // Sanity check:  For file and shm spaces, this should be the only	
    // mapping remaining.
    if ((type == AS_TYPE_FILE) || (type == AS_TYPE_SHM)) {
      if (canonical_page->get_mapping_count() != 1) {
	fprintf(stderr,
		"WARNING: Address_Space::clear_all_mappings(): ");
	fprintf(stderr,
		"Making zombie %s page with mapping count %d != 1\n",
		(type == AS_TYPE_FILE ? "file" : "shm"),
		canonical_page->get_mapping_count());
	canonical_page->make_zombie();
      }
    }

    // Sanity check: Each canonical page should have the correct type
    // for this address space.
    switch (canonical_page->get_type()) {
			
    case PAGE_TYPE_ANONYMOUS:
      assert(type == AS_TYPE_VIRTUAL);
      break;
				
    case PAGE_TYPE_FILE:
    case PAGE_TYPE_UNCACHEABLE:
      assert((type == AS_TYPE_VIRTUAL) || (type == AS_TYPE_FILE));
      break;
				
    case PAGE_TYPE_SHM:
      assert((type == AS_TYPE_VIRTUAL) || (type == AS_TYPE_SHM));
      break;
				
    default:
      fprintf(stderr, "Address_Space::clear_all_mappings(): ");
      fprintf(stderr, "Unknown canonical page file type\n");
      exit(1);
				
    }
		
    // Delete the page and the mapping entry to it and move to the
    // next page.
    virtual_to_canonical_map_t::iterator moribund = i;
    i++;
    map.erase(moribund);
    canonical_page->decrement_mapping_count();
    delete_if_unused_page(&canonical_page);
		
  }
	
} // Address_Space::clear_address_space
// ===================================================================



// ===================================================================
Address_Space*
create_address_space (const address_space_type_t type) {

  Address_Space* address_space = new Address_Space(type);
  if (address_space == NULL) {
    fprintf(stderr, "create_address_space(address_space_type_t): ");
    fprintf(stderr, "Failed allocation\n");
    exit(1);
  }
	
  return address_space;

} // create_address_space (address_space_type_t)
// ===================================================================



// ===================================================================
Address_Space*
create_address_space (const file_type_t type) {

  address_space_type_t address_space_type;
  switch (type) {

  case FILE_TYPE_REGULAR:
    address_space_type = AS_TYPE_FILE;
    break;

  case FILE_TYPE_PIPE:
  case FILE_TYPE_DEVICE:
  case FILE_TYPE_SOCKET:
    address_space_type = AS_TYPE_UNCACHED;
    break;

  default:
    fprintf(stderr, "Address_Space::create_address_space: ");
    fprintf(stderr,
	    "Unexpected file type specifier %c(%d)\n",
	    type,
	    (int)type);
    exit(1);

  }
		
  return create_address_space(address_space_type);

} // create_address_space (file_type_t)
// ===================================================================
