/**
 * \file Address_Space.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date May 2005
 * \brief The header that defines the Address_Space object type.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_ADDRESS_SPACE_HH)
#define _ADDRESS_SPACE_HH
// ===================================================================



// ===================================================================
// INCLUDES
#include <hash_map.h>
#include <stdio.h>
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
// TYPES

/**
 * \brief The type of an address space (virtual, file, shm, etc.).
 *
 * Each address space is either virtual, a file, or a shared memory
 * segment.  Entities accessed through the virtual file system such as
 * pipes, devices, or sockets are not cached, and so although we
 * assign an address space for each to maintain bookkeeping, we do not
 * pretend that the data associated with these entities is cached.
 */
typedef enum {
  AS_TYPE_VIRTUAL,
  AS_TYPE_FILE,
  AS_TYPE_SHM,
  AS_TYPE_UNCACHED
} address_space_type_t;

/** \brief A structure for comparing virtual page numbers used by
 *         tables indexed by virtual page number. */
struct equal_virtual_page_ID_s {
  bool operator() (const virtual_page_ID_t x,
		   const virtual_page_ID_t y) const {
    return (x == y);
  }
};

/** \brief A table type used to map virtual page IDs to canonical
 *         pages. */
typedef hash_map<virtual_page_ID_t,
		 Canonical_Page*,
		 hash<virtual_page_ID_t>,
		 equal_virtual_page_ID_s> virtual_to_canonical_map_t;
// ===================================================================



// ===================================================================
/**
 * \brief An object that describes a complete address space.
 *
 * An ``address space'' is any range of addresses that can be linearly
 * addressed as a sequence of pages.  It provides methods for
 * establishing, retrieving, and destroying mappings.
 **/
class Address_Space {
	
public: //METHODS
	
  /**
   * \brief The constructor.
   * \param type The type of the address space (virtual, file, shm,
   *        uncached).
   **/
  Address_Space (const address_space_type_t type) {
    this->type = type;
    this->size = 0;
  }
	
  /**
   * \brief The destructor.
   *
   * Clear the address space before disappearing.
   **/
  virtual ~Address_Space () {
    clear_all_mappings();
  }
	
  /**
   * \brief Lookup the canonical page for a given virtual page.
   * \param virtual_page_ID The virtual page number within this
   *        address space.
   * \return The canonical page to which the virtual page is mapped.
   *
   * Given a virtual page number within this address space, lookup the
   * canonical page to which it is mapped.  If no such mapping exists,
   * return NULL.
   **/
  Canonical_Page* lookup_canonical_page
  (virtual_page_ID_t virtual_page_ID);
	
  /**
   * \brief Lookup the canonical page for a given virtual page
   *        <i>safely</i>.
   * \param virtual_page_ID The virtual page number within this
   *        address space.
   * \return The canonical page to which the virtual page is mapped.
   *
   * Given a virtual page number within this address space, lookup the
   * canonical page to which it is mapped.  If no such mapping exists,
   * <i>abort processing</i>.
   **/
  Canonical_Page* safe_lookup_canonical_page
  (virtual_page_ID_t virtual_page_ID);
	
  /**
   * \brief Map a virtual page to a canonical page.
   * \param virtual_page_ID The virtual page number from which to map.
   * \param canonical_page The canonical page to which to map.
   *
   * Given a virtual page ID and a canonical page, map the former to
   * the latter.  For uncacheable spaces, perform no mapping, since
   * the mapping would not exist.
   **/
  void insert_mapping (const virtual_page_ID_t virtual_page_ID,
		       Canonical_Page* canonical_page);

  /**
   * \brief Remove a virtual->canonical mapping.
   * \param virtual_page_ID The virtual page number whose mapping to
   *        remove.
   **/
  void remove_mapping (const virtual_page_ID_t virtual_page_ID);

  /**
   * \brief Potentially expand the known size of a file's space.
   * \param size The number of pages for this address space.
   *
   * If the provided size indicates that this address space is larger
   * than previously believed, update the size.  Note that such space
   * tacking is only applied to file-based address spaces.
   **/
  void expand_size (file_offset_t size);

  /**
   * \brief Contract the known size of a file's space.
   * \param size The number of pages for this address space.
   *
   * Adopt the newly provide, shorter file size for this address
   * space.  Note that such space tacking is only applied to
   * file-based address spaces.
   **/
  void contract_size (file_offset_t size);

  /**
   * \brief Return the size of an address space.
   *
   * Note that sizes are only relevant to files, and so this method
   * aborts processing for any other type of space.
   **/
  file_offset_t get_size () const;

  /**
   * \brief Provide the type of this address space.
   * \return Return that the address space is either for a file, shm,
   *         virtual, or uncached.
   **/
  address_space_type_t get_type () const;
	
  /**
   * \brief Test that the address space type is as expected.
   * \param address_space_type The address space type to test.
   *
   * The address space type of this address space must match the given one.
   * If it does not, abort processing.
   **/
  void test_type (const address_space_type_t address_space_type) const;
	
  /**
   * \brief Test that the address space has the correct type for a
   *        given file type.
   * \param file_type The file type that is the basis of this address
   *        space.
   *
   * Based on a given file type, ensure that this address space is of
   * the right kind.  If it is not, abort processing.  Note that if
   * the given file type is FILE_TYPE_UNKNOWN, then no comparison is
   * made.
   **/
  void test_file_type (const file_type_t file_type) const;
		
  /**
   * \brief Unmap every virtual page in an address space.
   * \see Consumer::delete_if_unused_page()
   *
   * Remove all virtual->canonical mappings.  If an unmapped canonical
   * page is anonymous and defunct, then delete it.  If an IPC shared
   * memory segment is unmapped and thus defunct, it is marked for
   * deletion.
   **/
  void clear_all_mappings ();
		
protected: // DATA MEMBERS
			
  address_space_type_t type; /**< The type of this address space. */
  file_offset_t size; /**< The size of the address space (pages). */
  virtual_to_canonical_map_t map; /**< The map of virtual pages to
				   *   canonical pages. */
		
};
// ===================================================================



// ===================================================================
// HELPER FUNCTIONS

/**
 * \brief Create a new address space <i>safely</i>.
 * \param type The type of the address space (virtual, file, shm, or
 *        uncached).
 * \return A pointer to the newly created address space.
 **/
Address_Space* create_address_space (const address_space_type_t type);

/**
 * \brief Create a new address space <i>safely</i>.
 * \param type The type of a file object for which a space needs to be
 *        created.
 * \return A pointer to the newly created address space.
 **/
Address_Space* create_address_space (const file_type_t type);
// ===================================================================



// ===================================================================
#endif // _ADDRESS_SPACE_HH
// ===================================================================
