// ===================================================================
// Basic_Consumer.cc
// Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
// June 2004

// A subclass of Consumer that emits a basic page-level reference
// trace for virtual memory and file system references, emitting
// canonical page IDs.  See the header file for more details.
// ===================================================================



// ===================================================================
// INCLUDES

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "Basic_Consumer.hh"
#include "Canonical_Page.hh"
// ===================================================================



// ===================================================================
Basic_Consumer::Basic_Consumer (Reader* const reader,
				const char* const output_file_pathname) :
  Consumer(reader) {

  // Open the output file, checking for success.
  // Attempt to open the output file.
  int fd = open(output_file_pathname,
		O_WRONLY | O_CREAT | O_TRUNC,
		S_IRUSR | S_IWUSR);

  // Ensure that it did open.
  if (fd == -1) {
    perror("Basic_Consumer::Basic_Consumer(): Failed file open");
    exit(1);
  }

  // Link the file descriptor to a compressed output stream.  Ask that
  // the stream be for (w)riting at the highest (9) compression level
  // accepting (f)iltered data.
  output_stream = gzdopen(fd, "wb9f");

  // Ensure that the linking succeeded.
  if (output_stream == NULL) {
    fprintf(stderr, "Basic_Consumer::Basic_Consumer(): ");
    fprintf(stderr, "Failed compression open\n");
    exit(1);
  }

} // Basic_Consumer::Basic_Consumer
// ===================================================================



// ===================================================================
Basic_Consumer::~Basic_Consumer () {

  // Close the output file, checking for success.
  assert(output_stream != NULL);
  int result = gzclose(output_stream);
  if (result != Z_OK) {
    fprintf(stderr, "Basic_Consumer::~Basic_Consumer(): ");
    fprintf(stderr, "Close failure\n");
  }

} // Basic_Consumer::~Basic_Consumer
// ===================================================================



// ===================================================================
void
Basic_Consumer::act_on_reference_record () {

  // If we're still waiting to see the first scheduling event (which
  // would tell us which task is active), then we cannot yet process
  // this record.
  assert(merge_phase != PHASE_PRE_REFERENCE);
  if (merge_phase != PHASE_PROCESSING_REFERENCES) {
    return;
  }

  // Grab the process's virtual address space.
  context_ID_t context_ID = reference_record->virtual_page.context_ID;
  virtual_page_ID_t virtual_page_ID =
    reference_record->virtual_page.page_ID;
  Process* process = safe_context_ID_to_process(context_ID);
  Address_Space* address_space = process->address_space;

  // Lookup the canonical page ID.
  Canonical_Page* canonical_page =
    address_space->safe_lookup_canonical_page(virtual_page_ID);
  canonical_page_ID_t canonical_page_ID = canonical_page->get_ID();

  // Emit the new record only if the uncacheable page was not found.
  if (canonical_page != &canonical_uncacheable_page) {

    assert(!canonical_page->is_zombie());

    int result = gzprintf(output_stream,
			  "%c %lx\n",
			  reference_record->tag,
			  canonical_page_ID);

    if (result < 0) {
      fprintf(stderr, "Basic_Consumer::act_on_reference_record(): ");
      fprintf(stderr, "Failed record write\n");
      exit(1);
    }

  }

} // act_on_reference_record
// ===================================================================



// ===================================================================
void
Basic_Consumer::act_on_kernel_record () {

  // Only emit a record for a file system read or write operation, and
  // emit those only if reference processing has begun.
  if ((merge_phase == PHASE_PROCESSING_REFERENCES) &&
      ((kernel_record->tag == TAG_FILE_READ) ||
       (kernel_record->tag == TAG_FILE_WRITE))) {

    // Grab the file's address space.
    Address_Space* address_space =
      safe_file_ID_to_AS(kernel_record->file_ID, FILE_TYPE_UNKNOWN);

    // Loop through the file pages referenced by this operation.
    for (virtual_page_ID_t virtual_page_ID =
	   kernel_record->starting_file_page;
	 virtual_page_ID < kernel_record->ending_file_page;
	 virtual_page_ID++) {

      // Lookup the canonical page ID.
      Canonical_Page* canonical_page =
	address_space->safe_lookup_canonical_page(virtual_page_ID);
      canonical_page_ID_t canonical_page_ID = canonical_page->get_ID();

      // Emit the new record only if the uncacheable page is not found.
      if (canonical_page != &canonical_uncacheable_page) {

	assert(!canonical_page->is_zombie());

	int result = gzprintf(output_stream,
			      "%c %lx\n",
			      kernel_record->tag,
			      canonical_page_ID);

	if (result < 0) {
	  fprintf(stderr, "Basic_Consumer::act_on_kernel_record(): ");
	  fprintf(stderr, "Failed record write\n");
	  exit(1);
	}

      }
    }
  }
} // act_on_kernel_record
// ===================================================================
