// ===================================================================
// Basic_Named_Consumer.cc
// Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
// March 2005

// A subclass of Basic_Consumer that emits a basic page-level
// reference trace for virtual memory and file system references,
// emitting canonical page IDs with tags that indicate dirtiness and
// the named/unnamed status of the page.  See the header file for more
// details.
// ===================================================================



// ===================================================================
// INCLUDES

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "Basic_Named_Consumer.hh"
#include "Canonical_Page.hh"
// ===================================================================



// ===================================================================
Basic_Named_Consumer::Basic_Named_Consumer
(Reader* const reader,
 const char* const output_file_pathname) :
  Basic_Consumer(reader, output_file_pathname) {

  // Nothing additional to do.

} // Basic_Named_Consumer::Basic_Named_Consumer
// ===================================================================



// ===================================================================
void
Basic_Named_Consumer::act_on_reference_record () {

  // If we're still waiting to see the first scheduling event (which
  // would tell us which task is active), then we cannot yet process
  // this record.
  assert(merge_phase != PHASE_PRE_REFERENCE);
  if (merge_phase != PHASE_PROCESSING_REFERENCES) {
    return;
  }

  // Grab the virtual->canonical map for the given context.
  Process* process =
    safe_context_ID_to_process(reference_record->virtual_page.context_ID);
  Address_Space* address_space = process->address_space;

  // Lookup the canonical page.
  virtual_page_ID_t page_ID = reference_record->virtual_page.page_ID;
  Canonical_Page* canonical_page =
    address_space->safe_lookup_canonical_page(page_ID);

  // Determine the tag to emit.  For onymous (file) pages, the tag
  // will stay the same; for anonymous (swap) pages, we change the tag
  // to record its anonymity.
  tag_t tag = reference_record->tag;
  if (canonical_page->get_type() != PAGE_TYPE_FILE) {

    switch (tag) {

    case TAG_SINGLE_READ:
      tag = TAG_SINGLE_ANONYMOUS_READ;
      break;

    case TAG_MULTIPLE_READ:
      tag = TAG_MULTIPLE_ANONYMOUS_READ;
      break;

    case TAG_SINGLE_WRITE:
      tag = TAG_SINGLE_ANONYMOUS_WRITE;
      break;

    case TAG_MULTIPLE_WRITE:
      tag = TAG_MULTIPLE_ANONYMOUS_WRITE;
      break;

    default:
      fprintf(stderr,
	      "Basic_Named_Consumer::act_on_reference_record(): ");
      fprintf(stderr,
	      "Unknown tag %c(%d) on anonymous page\n",
	      tag,
	      (int)tag);
      exit(1);

    }

  }

  // Emit the new record, but not for the uncacheable page.
  if (canonical_page != &canonical_uncacheable_page) {

    int tagResult = gzwrite(output_stream, &tag, sizeof(tag_t));
    virtual_page_ID_t canonical_ID = canonical_page->get_ID();
    int pageIDResult = gzwrite(output_stream,
			       &canonical_ID,
			       sizeof(canonical_ID));

    if ((tagResult == 0) || (pageIDResult == 0)) {
      fprintf(stderr, "Basic_Consumer::act_on_reference_record(): ");
      fprintf(stderr, "Failed record write\n");
      exit(1);
    }

  }

} // act_on_reference_record
// ===================================================================
