/**
 * \file Canonical_Page.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date May 2005
 * \brief The code for the Canonical_Page class.
 *
 * Defines the methods within a Canonical_Page object, which itself
 * represents the underlying, unique page for each virtual, file, and
 * shm page within the canonical address space
 **/



// ===================================================================
// INCLUDES

#include "Canonical_Page.hh"
// ===================================================================



// ===================================================================
Canonical_Page::Canonical_Page (const page_type_t type) {

  // File and shm pages should be created with one of the other
  // constructors.
  assert((type != PAGE_TYPE_FILE) && (type != PAGE_TYPE_SHM));

  // Assign the fields.
  assign_next_ID();
  mapping_count = 0;
  this->type = type;
  virtual_ID = 0;
  zombie = false;

} // Canonical_Page::Canonical_Page (page_type_t)
// ===================================================================



// ===================================================================
Canonical_Page::Canonical_Page (const file_ID_s& file_ID,
				const virtual_page_ID_t virtual_ID) {

  // Assign the fields.
  assign_next_ID();
  mapping_count = 0;
  type = PAGE_TYPE_FILE;
  this->file_ID = file_ID;
  this->virtual_ID = virtual_ID;
  zombie = false;

} // Canonical_Page::Canonical_Page (file_ID_s&, virtual_page_ID_t)
// ===================================================================



// ===================================================================
Canonical_Page::Canonical_Page (const shm_ID_t shm_ID,
				const virtual_page_ID_t virtual_ID) {

  // Assign the fields.
  assign_next_ID();
  mapping_count = 0;
  type = PAGE_TYPE_SHM;
  this->shm_ID = shm_ID;
  this->virtual_ID = virtual_ID;
  zombie = false;

} // Canonical_Page::Canonical_Page (shm_ID_t, virtual_page_ID_t)
// ===================================================================



// ===================================================================
void
Canonical_Page::assign_next_ID () {

  static canonical_page_ID_t next_ID = 1;

  // Have we exhausted the canonical address space (that is, rolled
  // over the page counter?)  If so, abort processing.
  assert(next_ID != 0);

  // Assign the next available ID.
  ID = next_ID++;

} // Canonical_Page::assign_next_ID
// ===================================================================



// ===================================================================
Canonical_Page*
create_canonical_page (const page_type_t type) {
	
  // Allocate a new canonical page and return it.
  Canonical_Page* page = new Canonical_Page(type);
  assert(page != NULL);
  assert(page->get_type() == type);
  return page;

} // create_canonical_page (page_type_t)
// ===================================================================



// ===================================================================
Canonical_Page*
create_canonical_page (const file_ID_s& file_ID,
		       const virtual_page_ID_t virtual_ID) {

  Canonical_Page* page = new Canonical_Page(file_ID, virtual_ID);
  assert(page != NULL);
  assert(page->get_type() == PAGE_TYPE_FILE);
  return page;

} // create_canonical_page (file_ID_s&, virtual_page_ID_t)
// ===================================================================



// ===================================================================
Canonical_Page*
create_canonical_page (const shm_ID_t shm_ID,
		       const virtual_page_ID_t virtual_ID) {

  Canonical_Page* page = new Canonical_Page(shm_ID, virtual_ID);
  assert(page != NULL);
  assert(page->get_type() == PAGE_TYPE_SHM);
  return page;

} // create_canonical_page (shm_ID_t, virtual_page_ID_t)
// ===================================================================



// ===================================================================
// Take a handle to the canonical page, making it possible to nullify
// the pointer to the page if it is deleted by this method.

void
delete_if_unused_page (Canonical_Page** const canonical_page_handle) {

  // Sanity check: This handle should lead us to a legitimate
  // canonical page.
  assert(canonical_page_handle != NULL);
  Canonical_Page* canonical_page = *canonical_page_handle;
  assert(canonical_page != NULL);
	
  // If this page is not mapped by any address space, then delete it.
  if (canonical_page->get_mapping_count() == 0) {
		
    delete canonical_page;
    canonical_page = NULL;
			
  }

} // delete_if_unused_page
// ===================================================================
