/**
 * \file Canonical_Page.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date May 2005
 * \brief The header that defines the Canonical_Page object type.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_CANONICAL_PAGE_HH)
#define _CANONICAL_PAGE_HH
// ===================================================================



// ===================================================================
// INCLUDES
#include <stdio.h>
#include "types-and-constants.hh"
#include "Address_Space.hh"
// ===================================================================



// ===================================================================
// TYPES

/** \brief The type for a marker that indicates the type of a page
 *         (anonymous, file mmap'ed, etc.) */
typedef enum {
  PAGE_TYPE_ANONYMOUS,
  PAGE_TYPE_FILE,
  PAGE_TYPE_SHM,
  PAGE_TYPE_UNCACHEABLE
} page_type_t;

/** \brief The type for a canonical page number. */
typedef unsigned long int canonical_page_ID_t;
// ===================================================================



// ===================================================================
/**
 * \brief An object that describes a canonical page.
 *
 * A canonical page is the fundamental, underlying page identity for
 * any virtual or file page.  It has its own identifying number within
 * the canonical page space.
 **/
class Canonical_Page {

public: // METHODS

  /** \brief The standard constructor. */
  Canonical_Page (const page_type_t type);

  /** \brief The file page constructor. */
  Canonical_Page (const file_ID_s& file_ID,
		  const virtual_page_ID_t virtual_ID);

  /** \brief The shm page constructor. */
  Canonical_Page (const shm_ID_t shm_ID,
		  const virtual_page_ID_t virtual_ID);

  /** \brief Get the page number within the canonical address
   *         space. */
  canonical_page_ID_t get_ID () const {
    return ID;
  }

  /** \brief Get the page's type. */
  page_type_t get_type () const {
    return type;
  }

  /** \brief Increment the mapping count. */
  void increment_mapping_count () {
    mapping_count++;
  }

  /** \brief Decrement the mapping count. */
  void decrement_mapping_count () {
    mapping_count--;
  }

  /** \brief Get the mapping count. */
  unsigned int get_mapping_count () const {
    return mapping_count;
  }

  /** \brief Get the file ID (for file pages). */
  file_ID_s get_file_ID () const {
    assert(type == PAGE_TYPE_FILE);
    return file_ID;
  }

  /** \brief Get the shm ID (for shm pages). */
  shm_ID_t get_shm_ID () const {
    assert(type == PAGE_TYPE_SHM);
    return shm_ID;
  }

  /** \brief Get the virtual page ID (for file and shm pages). */
  virtual_page_ID_t get_virtual_page_ID () const {
    assert((type == PAGE_TYPE_FILE) || (type == PAGE_TYPE_SHM));
    return virtual_ID;
  }

  /**
   * \brief Mark the page as a zombie.
   *
   * Some file and shm pages <i>should</i> be deleted when their home
   * address spaces are themselves deleted, but sometimes (for unknown
   * reasons), they contain mapping counts that indicate that other
   * space(s) map them.  Thus, we mark them zombies to detect later
   * uses that indicate a more troublesome error.
   **/
  void make_zombie () {
    zombie = true;
  }

  /** \brief Return whether this page is a zombie. */
  bool is_zombie () const {
    return zombie;
  }

protected: // METHODS

  /**
   * \brief Assign the next available canonical page ID.
   *
   *  Assign the next, linearly assign canonical page ID.  Note that
   *  these IDs are never reused.  Test for canonical address space
   *  exhaustion and abort processing if it has occurred.
   **/
  void assign_next_ID ();

protected: // DATA MEMBERS

  /** \brief The page number in canonical space. */
  canonical_page_ID_t ID;

  /** \brief The number of virtual mappings to this page. */ 
  unsigned int mapping_count;

  /** \brief The type of this page (anonymous, file, shm,
   *         uncacheable). */
  page_type_t type;

  union {

    /** \brief The file ID (for file pages) of the owner space. */
    file_ID_s file_ID;

    /** \brief The shm ID (for shm pages) of the owner space. */
    shm_ID_t shm_ID;

  };

  /** \brief The offset into the home address space of the page (for
   *         file and shm pages). */
  virtual_page_ID_t virtual_ID;

  /** \brief Whether the page should have been deleted but was not due
   *         to the mapping count. */
  bool zombie;

}; // class Canonical_Page
// ===================================================================



// ===================================================================
// HELPER FUNCTIONS

/**
 * \brief Create a new canonical page of a given type <i>safely</i>.
 * \param type The type of the canonical page (virtual, file, shm, or
 *        uncacheable).
 * \return A pointer to the newly created canonical page.
 **/
Canonical_Page* create_canonical_page (const page_type_t type);

/**
 * \brief Create a new canonical file page <i>safely</i>.
 * \param file_ID The ID of the file to which this page will belong.
 * \param virtual_ID The virtual page number within the file.
 * \return A pointer to the new created canonical page.
 **/
Canonical_Page* create_canonical_page (const file_ID_s& file_ID,
				       const virtual_page_ID_t virtual_ID);

/**
 * \brief Create a new canonical shm page <i>safely</i>.
 * \param shm_ID The ID of the shm segment to which this page will
 *        belong.
 * \param virtual_ID The virtual page number within the file.
 * \return A pointer to the new created canonical page.
 **/
Canonical_Page* create_canonical_page (const shm_ID_t shm_ID,
				       const virtual_page_ID_t virtual_ID);

/**
 * \brief Delete a canonical page if it is defunct.
 * \param canonical_page_handle The canonical page to examine and
 *        possibly delete.
 *
 * If the given canonical page is no longer mapped into any address
 * space, delete it.  Note that a handle is passed so that the
 * caller's pointer can be cleared if the object is deleted.
 **/
void delete_if_unused_page
(Canonical_Page** const canonical_page_handle);
// ===================================================================



// ===================================================================
#endif // _CANONICAL_PAGE_HH
// ===================================================================
