/**
 * \file Consumer.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date June 2004
 * \brief Code of base class for reading traces and rebuilding system
 *        state.
 *
 * A class used to read both a kernel and reference trace and, with
 * each record, update data structures that reflect the state of the
 * processes and the memory.
 *
 * Note that this is a PURE VIRTUAL CLASS: It does not define the
 * act_on_kernel_record(), act_on_reference_record(), and clean_up()
 * methods.  By defining these methods, a subclass can choose to
 * maintain extra data structures or to produce output before
 * processing, at the processing of each record, and after processing.
 **/



// ===================================================================
// INCLUDES

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Consumer.hh"
#include "Canonical_Page.hh"
// ===================================================================



// ===================================================================
Consumer::Consumer (Reader* const reader) {
	
  // Hold a pointer to the reference reader.
  this->reader = reader;
	
  // Initialize some data members.
  next_process_ID = KERNEL_PROCESS_ID + 1;
  merge_phase = PHASE_PRE_REFERENCE;

  // Create space to hold one reference record and one kernel record.
  reference_record = new reference_record_s;
  kernel_record = new kernel_record_s;
  if ((reference_record == 0) || (kernel_record == 0)) {
    fprintf(stderr, "Consumer(): Unable to allocate record space\n");
    exit(1);
  }

  // Allocate and initialize the the handle dispatch table, then fill
  // it with the correct entries.
  handler_dispatch_table = new handler_t[NUMBER_POSSIBLE_TAG_VALUES];
  if (handler_dispatch_table == NULL) {
    fprintf(stderr, "Consumer::Consumer(): ");
    fprintf(stderr, "Could not allocate handler dispatch table\n");
    exit(1);
  }
  for (int i = 0; i < NUMBER_POSSIBLE_TAG_VALUES; i++) {
    handler_dispatch_table[i] = &Consumer::handle_invalid;
  }
	
  assign_handler(TAG_SCHEDULE, &Consumer::handle_schedule);
  assign_handler(TAG_FORK, &Consumer::handle_fork);
  assign_handler(TAG_EXEC, &Consumer::handle_exec);
  assign_handler(TAG_EXIT, &Consumer::handle_exit);
  assign_handler(TAG_CONTEXT_ASSIGNMENT, &Consumer::handle_context_assignment);
  assign_handler(TAG_DUPLICATE_RANGE, &Consumer::handle_duplicate_range);
  assign_handler(TAG_MMAP_FILE, &Consumer::handle_file_mmap_range);
  assign_handler(TAG_MMAP_ANONYMOUS, &Consumer::handle_anonymous_mmap_range);
  assign_handler(TAG_MREMAP, &Consumer::handle_mremap);
  assign_handler(TAG_MUNMAP, &Consumer::handle_munmap_range);
  assign_handler(TAG_COMPLETE_UNMAP, &Consumer::handle_complete_unmap);
  assign_handler(TAG_COW_UNMAP, &Consumer::handle_cow_unmap);
  assign_handler(TAG_SHMAT, &Consumer::handle_shmat);
  assign_handler(TAG_SHM_DESTROY, &Consumer::handle_shm_destroy);
  assign_handler(TAG_FILE_OPEN, &Consumer::handle_file_open);
  assign_handler(TAG_FILE_CLOSE, &Consumer::handle_no_op);
  assign_handler(TAG_FILE_READ, &Consumer::handle_file_read_or_write);
  assign_handler(TAG_FILE_WRITE, &Consumer::handle_file_read_or_write);
  assign_handler(TAG_FILE_DELETE, &Consumer::handle_file_delete);
  assign_handler(TAG_FILE_TRUNCATE, &Consumer::handle_file_truncate);
  assign_handler(TAG_ACCEPT, &Consumer::handle_no_op);
  assign_handler(TAG_INVALID, &Consumer::handle_invalid);
  assign_handler(TAG_END_OF_TRACE, &Consumer::handle_invalid);
  assign_handler(TAG_DEBUG, &Consumer::handle_cow_unmap);
	
  // Create the input buffer.
  buffer = new char[BUFFER_SIZE];
  if (buffer == NULL) {
    fprintf(stderr, "Consumer::Consumer(): ");
    fprintf(stderr, "Unable to allocation buffer\n");
    exit(1);
  }
	
} // Consumer::Consumer
// ===================================================================



// ===================================================================
void
Consumer::assign_handler (tag_t tag, handler_t handler) {
	
  int index = (int)tag;
  if (index >= NUMBER_POSSIBLE_TAG_VALUES) {
    fprintf(stderr,
	    "Consumer::assign_handler():  Invalid tag/index = %c/%d >= %d\n",
	    tag,
	    index,
	    NUMBER_POSSIBLE_TAG_VALUES);
    exit(1);
  }
  assert(handler_dispatch_table != NULL);
  handler_dispatch_table[index] = handler;
	
}
// ===================================================================



// ===================================================================
Consumer::~Consumer () {
	
  // C: We do nothing here to clean up this object.  We (lazily)
  // assume that when this object dies, the process is complete
  // anyway.
	
} // Consumer::~Consumer
// ===================================================================



// ===================================================================
// Perform the merging.
void
Consumer::merge () {
	
  // Complete initialization.
  initialize();

  // Read the first kernel and reference records to be processed.
  reader->read_kernel_record(kernel_record);
  reader->read_reference_record(reference_record);
	
  // Merge the traces until the last reference has been processed.
  while (reference_record->tag != TAG_END_OF_TRACE) {
		
    // It should be impossible for the timestamps of the current
    // kernel and reference records to match.
    if (kernel_record->cycle_timestamp ==
	reference_record->cycle_timestamp) {
			
      fprintf(stderr, "Consumer::merge(): Identical timestamps\n");
      fprintf(stderr,
	      "  kernel_record->cycle_timestamp = %qx at record %qd\n",
	      kernel_record->cycle_timestamp,
	      kernel_record->record_number);
      fprintf(stderr,
	      "  reference_record->cycle_timestamp = %qx at record %qd\n",
	      reference_record->cycle_timestamp,
	      reference_record->record_number);
      exit(1);
			
    }
		
    // Which trace contains the next record?
    if (kernel_record->cycle_timestamp <
	reference_record->cycle_timestamp) {
			
      // The kernel trace record is next. Update the time that has
      // passed since the last record for the scheduled task.
      //
      // Special case: The previous record may have terminated the
      // scheduled task, thus obviating the need to mark its time.
      // However, be sure that the current record is a scheduling
      // record that will set a new scheduled task.
      if (scheduled_task == NULL) {
				
	assert(kernel_record->tag == TAG_SCHEDULE);
				
      } else {
				
	scheduled_task->mark_time(kernel_record->cycle_timestamp,
				  kernel_record->reference_timestamp);
				
      }
			
      // Update the maps with this kernel record, and act on the
      // record (that is, output and other bookkeeping).
      update_with_kernel_record();
      act_on_kernel_record();
			
      // Read the next kernel record.
      reader->read_kernel_record(kernel_record);
      
    } else {
			
      // The reference record is processed next.  Update the time that
      // has passed since the last record for the scheduled task.
      scheduled_task->mark_time(reference_record->cycle_timestamp,
				reference_record->reference_timestamp);
			
      // Update the maps with this reference record, and act on the
      // record (that is, output and other bookkeeping).
      update_with_reference_record();
      act_on_reference_record();
			
      // Read the next reference record.
      reader->read_reference_record(reference_record);
			
    }
		
  } // Record reading loop.
	
  // Complete processing by finishing output and other bookkeeping.
  clean_up();
	
} // Consumer::merge
// ===================================================================


// ===================================================================
Task*
Consumer::safe_task_ID_to_task (const task_ID_t task_ID) {
	
  // Attempt lookup.
  task_ID_to_Task_map_t::iterator i = task_ID_to_Task_map.find(task_ID);
	
  // If the lookup failed then abort.
  if (i == task_ID_to_Task_map.end()) {
    fprintf(stderr, "Consumer::safe_task_ID_to_task(): ");
    fprintf(stderr,
	    "Failed lookup at record %qd\n",
	    kernel_record->record_number);
    exit(1);
  }
	
  // Return the lookup result.
  return (*i).second;
	
}
// ===================================================================



// ===================================================================
Process*
Consumer::safe_context_ID_to_process (const context_ID_t context) {
	
  // Attempt lookup.
  context_ID_to_Process_map_t::iterator i =
    context_ID_to_Process_map.find(context);
	
  // If the lookup failed then abort.
  if (i == context_ID_to_Process_map.end()) {
    fprintf(stderr, "Consumer::safe_context_ID_to_process(): ");
    fprintf(stderr, "Failed lookup\n");
    exit(1);
  }
	
  // Return the lookup result.
  return (*i).second;
	
}
// ===================================================================



// ===================================================================
bool
Consumer::safe_live (const task_ID_t task_ID) {
	
  // Attempt lookup.
  live_map_t::iterator i = live_map.find(task_ID);
	
  // If the lookup failed then abort.
  if (i == live_map.end()) {
    fprintf(stderr, "safe_live(): Failed lookup\n");
    exit(1);
  }
	
  // Return the lookup result.
  return (*i).second;
	
}
// ===================================================================



// ===================================================================
Address_Space*
Consumer::safe_file_ID_to_AS (const file_ID_s& file_ID,
			      const file_type_t file_type) {
	
  Address_Space* address_space = NULL;
	
  // Attempt to lookup the address space.
  file_ID_to_AS_map_t::iterator i = file_ID_to_AS_map.find(file_ID);
	
  // Did the lookup succeed?
  if (i == file_ID_to_AS_map.end()) {
		
    // No.  Create an address space of the needed type.
    address_space = create_address_space(file_type);
		
    // Associate the given file ID with this new address space.
    file_ID_to_AS_map[file_ID] = address_space;
		
  } else {
		
    // The address space was found.  Sanity check that it is the
    // expected type (if there was any expectation).
    address_space = (*i).second;
    address_space->test_file_type(file_type);
		
  }
	
  // Return the lookup result.
  return address_space;
	
}
// ===================================================================



// ===================================================================
Address_Space*
Consumer::safe_shm_ID_to_AS (const shm_ID_t& shm_ID) {
	
  Address_Space* address_space = NULL;
	
  // Attempt lookup.
  shm_ID_to_AS_map_t::iterator i = shm_ID_to_AS_map.find(shm_ID);
	
  // Did the lookup succeed?
  if (i == shm_ID_to_AS_map.end()) {
		
    // No.  Allocate a new address space, and ensure that the
    // allocation succeeded.
    address_space = create_address_space(AS_TYPE_SHM);
    if (address_space == NULL) {
      fprintf(stderr, "safe_shm_ID_to_AS_map(): Failed allocation\n");
      exit(1);
    }
		
    // Associate the given shm ID with this new address space.
    shm_ID_to_AS_map[shm_ID] = address_space;
		
  } else {
		
    // Yes, use the address space found via lookup.  Sanity check that
    // it is of the correct type.
    address_space = (*i).second;
    address_space->test_type(AS_TYPE_SHM);
		
  }
	
  // Return the lookup result.
  return address_space;
	
} // safe_shm_ID_to_AS_map
// ===================================================================



// ===================================================================
void
Consumer::delete_task (Task* task) {
	
  assert(task != NULL);
  delete task;
	
} // Consumer::delete_task
// ===================================================================



// ===================================================================
// Take a handle to the process, making it possible to nullify the
// pointer to the process if it is deleted by this method.

void
Consumer::delete_if_defunct_process (Process** const process_handle) {
	
  Process* process = *process_handle;
	
  // Are there any tasks left for this process?
  if (process->task_list.empty()) {
		
    // Find and erase the context_ID->process mapping.
    context_ID_to_Process_map_t::iterator i =
      context_ID_to_Process_map.find(process->context_ID);
    if (i != context_ID_to_Process_map.end()) {
			
      context_ID_to_Process_map.erase(i);
			
    }
		
    // Delete the address space and the process object.
    delete process->address_space;
    delete process;
    *process_handle = 0;
		
  } // Process has no remaining tasks
	
} // Consumer::delete_if_defunct_process
// ===================================================================



// ===================================================================
void
Consumer::unlink_task_and_process (Task* const task) {
	
  // Grab a pointer the task's process.
  Process* process = task->process;
	
  // If the task is not associated with a process, then there's no
  // work to do here.
  if (process == NULL) {
    return;
  }
	
  // Clear the process pointer in the task object, thus performing
  // the first half of the unlinking.
  task->process = NULL;
	
  // Search through the process's task list for this task.
  list<Task*>::iterator i = process->task_list.begin();
  bool done = false;
  while (!done) {
		
    // Sanity check:  We should never reach the end of the list
    // without finding the task.
    if (i == process->task_list.end()) {
      fprintf(stderr, "Consumer::unlink_task_and_process(): ");
      fprintf(stderr, "Task not found in process\n");
      exit(1);
    }
		
    // If we've found it, delete it and end the search.  Otherwise,
    // keep searching.
    if (*i == task) {
			
      // It is.  Erase it from the list, thus performing the second
      // half of the unlinking.
      process->task_list.erase(i);
      done = true;
			
    } else {
			
      // It is not.  Continue searching.
      i++;
			
    }
  } // Loop over task IDs in the process
	
  // The process may be defunct as a result of the unlinking, and if
  // so, it should be deleted.
  delete_if_defunct_process(&process);
	
} // Consumer::unlink_task_and_process
// ===================================================================



// ===================================================================
void
Consumer::handle_schedule () {
	
  // Find the task object based on the given ID.
  scheduled_task =
    safe_task_ID_to_task(kernel_record->task_ID);
	
  // Mark the start of the quanta for the scheduled task.  Note that
  // reference time, provided as a virtual value, cannot have changed
  // for this task, but that cycle time, provided as a global value,
  // will have.
  scheduled_task->mark_quanta_start(kernel_record->cycle_timestamp);
	
  // If this is the first scheduling event since the first observed
  // reference, mark the change of phase to processing references.
  if (merge_phase == PHASE_WAITING_FOR_SCHEDULE) {
    merge_phase = PHASE_PROCESSING_REFERENCES;
  }
	
} // Consumer::handle_schedule
// ===================================================================



// ===================================================================
void
Consumer::handle_fork () {
	
  // Sanity check: There should be no task already using the given ID.
  task_ID_to_Task_map_t::iterator i =
    task_ID_to_Task_map.find(kernel_record->task_ID);
  if (i != task_ID_to_Task_map.end()) {
    fprintf(stderr, "Consumer::handle_fork(): Duplicate task ID\n");
    exit(1);
  }
	
  // Create a new task object.
  Task* child_task = allocate_new_task();
	
  // Look up the parent task.
  Task* parent_task =
    safe_task_ID_to_task(kernel_record->parent_task_ID);
	
  // Assign the new task's fields.  Initially, the task has the same
  // context as the parent.
  child_task->task_ID = kernel_record->task_ID;
  child_task->parent_task = parent_task;
  assert(parent_task->process != NULL);
  parent_task->process->task_list.push_front(child_task);
  child_task->process = parent_task->process;
	
  // Create a table entry mapping the task ID to the new task object.
  task_ID_to_Task_map[kernel_record->task_ID] = child_task;
	
} // Consumer::handle_fork
// ===================================================================



// ===================================================================
void
Consumer::handle_exec () {
	
  // Grab the task process associated with this task.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  Process* process = task->process;
	
  // Assign the new executable image.
  process->image_file_ID = kernel_record->file_ID;
  strcpy(process->image_pathname, kernel_record->filename);
	
} // Consumer::handle_exec
// ===================================================================



// ===================================================================
void
Consumer::handle_exit () {
	
  // Lookup the task and process objects for the given task ID.
  task_ID_to_Task_map_t::iterator task_iterator =
    task_ID_to_Task_map.find(kernel_record->task_ID);
  if (task_iterator == task_ID_to_Task_map.end()) {
    fprintf(stderr, "Consumer::handle_exit(): Failed lookup\n");
    exit(1);
  }
  Task* task = (*task_iterator).second;
	
  // Separate the task and its process.  If the process becomes
  // defunct as a result, it will be deleted.
  unlink_task_and_process(task);
	
  // Clear the task_ID->process mapping.
  task_ID_to_Task_map.erase(task_iterator);
	
  // Delete the task.  If this is the currently scheduled task, record
  // that fact too.
  if (task == scheduled_task) {
    scheduled_task = NULL;
  }
  delete_task(task);
  task = NULL;
	
} // Consumer::handle_exit
// ===================================================================



// ===================================================================
void
Consumer::handle_context_assignment () {
	
  // Grab the task object for the given task ID.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
	
  // Unlink this task from its current process (if any).
  unlink_task_and_process(task);
	
  // Get the mapping from the given context ID to its process.  Is it
  // mapped to any process?
  context_ID_to_Process_map_t:: iterator process_iterator =
    context_ID_to_Process_map.find(kernel_record->context_ID);
  Process* process;
  if (process_iterator == context_ID_to_Process_map.end()) {
		
    // The context ID is not yet associated with a process.  Allocate
    // a new process object.
    process = allocate_new_process();
		
    // Assign the process a new ID.
    process->process_ID = next_process_ID++;
    if (process->process_ID == 0) {
      fprintf(stderr, "Consumer::handle_context_assignment(): ");
      fprintf(stderr, "Exhausted process ID space\n");
      exit(1);
    }
		
    // Create a new address space.
    process->address_space = create_address_space(AS_TYPE_VIRTUAL);
		
    // Assign the given context ID to the process.
    process->context_ID = kernel_record->context_ID;
		
    // Assign a mapping from this context ID to this process.
    context_ID_to_Process_map[process->context_ID] = process;
		
    // Assign the process an executable image by copying it from this
    // task's parent's process.
    Process* parent_process = task->parent_task->process;
    process->image_file_ID = parent_process->image_file_ID;
    strcpy(process->image_pathname, parent_process->image_pathname);
		
  } else {
		
    // A process already exists using the given context, so this task
    // is joining an existing process.  Grab that process.
    process = (*process_iterator).second;
		
    // Sanity check:  This process should explicitly indicate that it
    // is using the given context ID.
    if (process->context_ID != kernel_record->context_ID) {
      fprintf(stderr, "Consumer::handle_context_assignment(): ");
      fprintf(stderr, "Process and map disagree on context ID\n");
      exit(1);
    }
		
  }
	
  // Link the task and the process to one another.
  task->process = process;
  process->task_list.push_front(task);
	
} // Consumer::handle_context_assignment
// ===================================================================



// ===================================================================
void
Consumer::handle_duplicate_range () {
	
  // Get the address spaces for both the original (parent) and
  // duplicate (child).
  Process* original_process =
    safe_context_ID_to_process(kernel_record->context_ID);
  Address_Space* original_space = original_process->address_space;
  Process* duplicate_process =
    safe_context_ID_to_process(kernel_record->duplicate_context_ID);
  Address_Space* duplicate_space = duplicate_process->address_space;
	
  // Traverse every virtual page in the range, making that page in the
  // duplicate context point to the same canonical page as that page
  // does in the original context.
  for (virtual_page_ID_t current_page =
	 kernel_record->starting_virtual_page;
       current_page < kernel_record->ending_virtual_page;
       current_page++) {
		
    // We should never be trying to duplicate pages in the kernel's
    // address range.
    assert (current_page < KERNEL_VIRTUAL_PAGE_OFFSET);

    // Attempt to lookup the canonical page underlying the original.
    // Note that it is possible for a task to map an anonymous
    // region without having referenced it yet, leaving some pages in
    // that region without assigned canonical pages.
    Canonical_Page* canonical_page =
      original_space->lookup_canonical_page(current_page);
		
    // Has a canonical page yet been assigned to this page in the
    // original context?
    if (canonical_page == NULL) {
			
      // No.  File and shm pages are immediately assigned canonical
      // pages, so this must be an anonymous page.  Map it to the zero
      // page.
      original_space->insert_mapping(current_page,
				     &canonical_zero_page);

    }
		
    // The current page in the duplicate space cannot be mapped to
    // anything -- an error has occurred if it is.
    assert(duplicate_space->lookup_canonical_page(current_page) == NULL);
		
    // Map this canonical page into the duplicate context.
    duplicate_space->insert_mapping(current_page, canonical_page);
		
  }

} // Consumer::handle_duplicate_range
// ===================================================================



// ===================================================================
void
Consumer::do_file_mmap_range (const task_ID_t task_ID,
			      const file_ID_s& file_ID,
			      const virtual_page_ID_t starting_virtual_page,
			      const virtual_page_ID_t ending_virtual_page,
			      const virtual_page_ID_t starting_file_page,
			      const virtual_page_ID_t ending_file_page,
			      const file_type_t file_type) {
	
  // Get the address spaces for the process and the file.
  Task* task = safe_task_ID_to_task(task_ID);
  Address_Space* virtual_space = task->process->address_space;
  Address_Space* file_space = safe_file_ID_to_AS(file_ID, file_type);
	
  // If necessary, update our notion of the file's length.
  file_space->expand_size(ending_file_page + 1);
	
  // Traverse each page in the range, establishing an entry that maps
  // the task virtual page to the file page.
  virtual_page_ID_t task_page = starting_virtual_page;
  virtual_page_ID_t file_page = starting_file_page;
  while (task_page < ending_virtual_page) {
		
    // The task page should never be in the kernel's page range.
    assert(task_page < KERNEL_VIRTUAL_PAGE_OFFSET);
		
    // We should never go off the end of the file page range.
    assert(file_page < ending_file_page);

    // Attempt to lookup the canonical page underlying this file page.
    Canonical_Page* canonical_page =
      file_space->lookup_canonical_page(file_page);
		
    // Has a canonical page yet been assigned to this file page?
    if (canonical_page == NULL) {

      // No.  Create a new canonical page and map it into the
      // file's space.
      canonical_page = create_canonical_page(file_ID, file_page);
      file_space->insert_mapping(file_page, canonical_page);

    } else {

      // Yes.  Sanity check: The existing page should be either a file
      // page that is associated with the correct file, or it should
      // be an uncacheable page (unassociated with any particular
      // file).
      switch (canonical_page->get_type())
	{

	case PAGE_TYPE_FILE:
	  
	  assert(canonical_page->get_file_ID() == file_ID);
	  assert(canonical_page->get_virtual_page_ID() == file_page);
	  break;

	case PAGE_TYPE_UNCACHEABLE:
	  break;

	default:
	  assert(false);

	}
			
    }
		
    // The task shouldn't already have a mapping for its virtual page.
    assert(virtual_space->lookup_canonical_page(task_page) == NULL);
		
    // Map this canonical page into the task's space.
    virtual_space->insert_mapping(task_page, canonical_page);
		
    // Move to the next page in each space.
    task_page++;
    file_page++;
		
  }
	
} // Consumer::do_file_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::handle_file_mmap_range () {
	
  // Simply call on do_file_mmap_range() to do the real work.
  do_file_mmap_range(kernel_record->task_ID,
		     kernel_record->file_ID,
		     kernel_record->starting_virtual_page,
		     kernel_record->ending_virtual_page,
		     kernel_record->starting_file_page,
		     kernel_record->ending_file_page,
		     kernel_record->file_type);
	
} // Consumer::handle_file_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::do_anonymous_mmap_range
(const task_ID_t task_ID,
 const virtual_page_ID_t starting_virtual_page,
 const virtual_page_ID_t ending_virtual_page) {
	
  // Grab the virtual address space for the process.
  Task* task = safe_task_ID_to_task(task_ID);
  Address_Space* virtual_space = task->process->address_space;
	
  // Traverse each page in the range, establishing an entry that maps
  // the task virtual page to the canonical ``zero page''.
  virtual_page_ID_t page = starting_virtual_page;
  while (page < ending_virtual_page) {
		
    // The task page should never be in the kernel's page range.
    assert(page < KERNEL_VIRTUAL_PAGE_OFFSET);
		
    // The task shouldn't already have a mapping for this page in its
    // space.
    assert(virtual_space->lookup_canonical_page(page) == NULL);
		
    // Map the canonical zero page into the task's space.  Note that
    // this mapping may not reflect the reality in the kernel, which
    // may not immediately map the virtual pages to any canonical
    // page.  However, it is safe to assume that it is mapped to the
    // zero page because any reads will require a mapping to that
    // page, and any writes will require a COW operation.  (See the
    // handling of reference records, where such COW operations are
    // performed, even in the absence of an explicit kernel COW
    // record.)
    virtual_space->insert_mapping(page, &canonical_zero_page);
		
    // Move to the next page.
    page++;
		
  }
} // Consumer::do_anonymous_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::handle_anonymous_mmap_range () {
	
  // Simply call on do_anonymous_mmap_range() to do the real work.
  do_anonymous_mmap_range(kernel_record->task_ID,
			  kernel_record->starting_virtual_page,
			  kernel_record->ending_virtual_page);
	
} // Consumer::handle_anonymous_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::do_shm_mmap_range (const task_ID_t task_ID,
			     const shm_ID_t shm_ID,
			     const virtual_page_ID_t starting_virtual_page,
			     const virtual_page_ID_t ending_virtual_page) {

  // Grab task's virtual address space and the shared memory segment's
  // address space.
  Task* task = safe_task_ID_to_task(task_ID);
  Address_Space* virtual_space = task->process->address_space;
  Address_Space* shm_space = safe_shm_ID_to_AS(shm_ID);

  // Traverse every page in the range, establishing an entry that maps
  // the task virtual page to the shared memory segment page.  Note
  // that shared memory segments are assumed to start at an offset of
  // zero.
  virtual_page_ID_t task_page = starting_virtual_page;
  virtual_page_ID_t shm_page = 0;
  while (task_page < ending_virtual_page) {

    // The task page should never be in the kernel's page range.
    assert(task_page < KERNEL_VIRTUAL_PAGE_OFFSET);

    // Has a canonical page yet been assigned to this shared memory
    // page?
    Canonical_Page* canonical_page =
      shm_space->lookup_canonical_page(shm_page);
    if (canonical_page == NULL) {

      // No.  Create a new canonical page, associate it with this new
      // shm segment, and map it into the shared memory segment's
      // space.
      canonical_page = create_canonical_page(shm_ID, shm_page);
      shm_space->insert_mapping(shm_page, canonical_page);

    } else {
			
      // Yes.  Sanity checks:  The canonical page should be the right
      // one of this shm segment.
      assert(canonical_page->get_type() == PAGE_TYPE_SHM);
      assert(canonical_page->get_shm_ID() == shm_ID);
      assert(canonical_page->get_virtual_page_ID() == shm_page);
			
    }
		
    // The task should not already have a mapping for this page in its
    // space.
    assert(virtual_space->lookup_canonical_page(task_page) == NULL);
		
    // Map this canonical page into the task's space.
    virtual_space->insert_mapping(task_page, canonical_page);
		
    // Move to the next pair of pages.
    task_page++;
    shm_page++;
		
  }

} // Consumer::do_shm_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::handle_shmat () {
	
  // Simply call on do_shm_mmap_range() to do the real work.
  do_shm_mmap_range(kernel_record->task_ID,
		    kernel_record->shm_ID,
		    kernel_record->starting_virtual_page,
		    kernel_record->ending_virtual_page);
	
} // Consumer::handle_shmat
// ===================================================================



// ===================================================================
void
Consumer::handle_shm_destroy () {
	
  // Grab the context for this shm segment.
  shm_ID_to_AS_map_t::iterator i =
    shm_ID_to_AS_map.find(kernel_record->shm_ID);
	
  // It may be the case that we've not encountered this shm segment
  // before, as it need not be attached to be destroyed by some
  // process.  There's only work to do if we've attached this shm
  // segment before.
  if (i != shm_ID_to_AS_map.end()) {
		
    Address_Space* address_space =
      safe_shm_ID_to_AS(kernel_record->shm_ID);
		
    // Delete the address space and its association with the given shm
    // ID.
    delete address_space;
    shm_ID_to_AS_map.erase(i);
    
  }
	
} // Consumer::handle_shm_destroy
// ===================================================================



// ===================================================================
// SFK: This method is horribly long and should, some day, when life
// becomes boring and there are no more important things to do, be
// properly divided.

void
Consumer::handle_mremap () {
	
  // Grab the task's virtual address space.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  Address_Space* address_space = task->process->address_space;
	
  // Grab the starting virtual page numbers for the old and new
  // ranges
  virtual_page_ID_t old_page = kernel_record->old_starting_virtual_page;
  virtual_page_ID_t new_page = kernel_record->starting_virtual_page;

  // Sanity check: Since this is a Linux-specific system call, it
  // should be the case that the remapping covers a single VMA.
  // Therefore, for file and shm AS's the canonical page's type and
  // underlying address space (e.g. file ID, shm ID) should not
  // change, and the offset within that space (for files and shms)
  // should increment regularly.  Copy the starting canonical page use
  // it to track progression through the address space.
  bool expectation = false;
  virtual_page_ID_t expected_virtual_ID = 0;
  Canonical_Page expected_page =
    *(address_space->safe_lookup_canonical_page(old_page));
  if ((expected_page.get_type() == PAGE_TYPE_FILE) ||
      (expected_page.get_type() == PAGE_TYPE_SHM)) {
    expectation = true;
    expected_virtual_ID = expected_page.get_virtual_page_ID();
  }

  // Iterate through both ranges for the length of the old range.  For
  // each page visited in both ranges, map the new virtual page to the
  // same canonical page as the old virtual page.  Then unmap the old
  // virtual page.
  while (old_page < kernel_record->old_ending_virtual_page) {
		
    // Sanity check: Be sure that we're staying within the ranges.
    assert(new_page < kernel_record->ending_virtual_page);
		
    // Sanity check: Neither the old nor new virtual pages should be
    // in the kernel's page reange.
    assert(old_page < KERNEL_VIRTUAL_PAGE_OFFSET);
    assert(new_page < KERNEL_VIRTUAL_PAGE_OFFSET);
		
    // Sanity check: The task should not already have a mapping for
    // the new page in its space.
    assert(address_space->lookup_canonical_page(new_page) == NULL);
		
    // Grab the underlying canonical page for the old virtual page.
    // Sanity check: The page must exist, and it must match our
    // expected canonical page.  Note that ``matching'' can only be
    // fully tested for file and shm pages because they have a unique,
    // source context and virtual page ID within that context.
    Canonical_Page* canonical_page =
      address_space->safe_lookup_canonical_page(old_page);
    assert(expected_page.get_type() == canonical_page->get_type());
    if (expectation) {

      switch (expected_page.get_type())
	{

	case PAGE_TYPE_FILE:
	  assert(expected_page.get_file_ID() ==
		 canonical_page->get_file_ID());
	  assert(expected_virtual_ID ==
		 canonical_page->get_virtual_page_ID());
	  break;

	case PAGE_TYPE_SHM:
	  assert(expected_page.get_shm_ID() ==
		 canonical_page->get_shm_ID());
	  assert(expected_virtual_ID ==
		 canonical_page->get_virtual_page_ID());
	  break;

	default:
	  // Nothing to do in other cases.
	  break;

	}
    }

    // Create a mapping from the new virtual page to the canonical
    // page, and then remove the old mapping.
    address_space->insert_mapping(new_page, canonical_page);
    address_space->remove_mapping(old_page);
		
    // Move to the next old and new virtual pages, as well as the next
    // expected canonical page (for file and shm pages).
    old_page++;
    new_page++;
    expected_virtual_ID++;
		
  }

  // Call on other functions to create the new mappings of additional
  // pages (that is, those beyond the length of the old range).
  switch (expected_page.get_type()) {

  case PAGE_TYPE_ANONYMOUS:

    do_anonymous_mmap_range(kernel_record->task_ID,
			    new_page,
			    kernel_record->ending_virtual_page);
    break;
			
  case PAGE_TYPE_FILE:
			
    // The expected page is the next canonical page whose source is
    // the file associated with the mapping range being extended.
    // Thus, we can obtain the file ID and page range within that
    // file.  Note that the range of file pages can only be derived
    // from the range of virtual pages to be mapped.
    do_file_mmap_range(kernel_record->task_ID,
		       expected_page.get_file_ID(),
		       new_page,
		       kernel_record->ending_virtual_page,
		       expected_virtual_ID,
		       (expected_virtual_ID +
			(kernel_record->ending_virtual_page - new_page)),
		       FILE_TYPE_UNKNOWN);
    break;
			
  case PAGE_TYPE_SHM:
			
    // The expected page is the next canonical page whose source is
    // the shm segment associated with the mapping range being
    // extended.  Thus, We can obtain the shm ID from the expected
    // page.
    do_shm_mmap_range(kernel_record->task_ID,
		      expected_page.get_shm_ID(),
		      expected_virtual_ID,
		      kernel_record->ending_virtual_page);
    break;
			
  default:
			
    fprintf(stderr, "Consumer::handle_mremap(): ");
    fprintf(stderr,
	    "Canonical page with unknown type %c(%d) at record %qd\n",
	    expected_page.get_type(),
	    (int)expected_page.get_type(),
	    kernel_record->record_number);
    exit(1);
			
  }
	
} // Consumer::handle_mremap
// ===================================================================



// ===================================================================
void
Consumer::handle_munmap_range () {
	
  // Grab the task's virtual address space.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  Address_Space* address_space = task->process->address_space;
	
  // Traverse every page in the range, eliminating any mappings found.
  virtual_page_ID_t page = kernel_record->starting_virtual_page;
  while (page < kernel_record->ending_virtual_page) {
		
    // Sanity check: The task page should never be in the kernel's
    // page range.
    assert(page < KERNEL_VIRTUAL_PAGE_OFFSET);

    // Remove the mapping if it exists.
    address_space->remove_mapping(page);
		
    // Move to the next page.
    page++;
		
  }

} // Consumer::handle_munmap_range
// ===================================================================



// ===================================================================
void
Consumer::handle_complete_unmap () {
	
  // Is the context ID mapped to a process?  It may not be if the
  // process with which it was associated became defunct and was
  // deleted, taking the address space with it.
  context_ID_to_Process_map_t::iterator i =
    context_ID_to_Process_map.find(kernel_record->context_ID);
  if (i != context_ID_to_Process_map.end()) {
		
    // Clear the address space.
    Process* process = (*i).second;
    process->address_space->clear_all_mappings();

  }
	
} // Consumer::handle_complete_unmap
// ===================================================================



// ===================================================================
void
Consumer::handle_cow_unmap () {
	
  virtual_page_ID_t page = kernel_record->starting_virtual_page;
	
  // Sanity check: The page should not be in the kernel's page range.
  assert(page < KERNEL_VIRTUAL_PAGE_OFFSET);
	
  // Grab the task's virtual address space.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  Address_Space* address_space = task->process->address_space;
	
  // Grab the page, which must exist.
  Canonical_Page* original_canonical_page =
    address_space->safe_lookup_canonical_page(page);
	
  // Sanity check: The original page cannot have been an IPC shared
  // memory page.
  assert(original_canonical_page->get_type() != PAGE_TYPE_SHM);
	
  // Sanity check: When COW-unmapping an anonymous page, it should
  // never be the case that the sharing count reaches zero -- the last
  // task to write to the COW page should get that original.
  //
  // C: We _do_ see this phenomenon.  Why?  Harmless (we think) to let
  // it go...
  if ((original_canonical_page->get_type() == PAGE_TYPE_ANONYMOUS) &&
      (original_canonical_page->get_mapping_count() == 1)) {
    fprintf(stderr, "WARNING: Consumer::handle_cow_unmap(): ");
    fprintf(stderr,
	    "Anonymous page orphaned at record %qd\n",
	    kernel_record->record_number);
  }
	
  // Reassign this virtual page to a new canonical page, thus
  // representing the copy that is made by the COW mechanism.  Whether
  // the original page was anonymous or from a file, the new copy must
  // be anonymous.
  address_space->remove_mapping(page);
  Canonical_Page* new_canonical_page =
    create_canonical_page(PAGE_TYPE_ANONYMOUS);
  address_space->insert_mapping(page, new_canonical_page);
	
} // Consumer::handle_cow_unmap
// ===================================================================



// ===================================================================
void
Consumer::handle_file_open () {

  // Get the address space for the file.
  Address_Space* address_space =
    safe_file_ID_to_AS(kernel_record->file_ID,
		       kernel_record->file_type);

  // C: The trace record's file length should not be shorter than our
  // file length.  If it is, then we missed a truncation (not likely)
  // or a deletion/creation (more likely).  This case does seem to
  // occur, so we handle it as though a truncation must have occurred;
  // by doing so, we prevent re-use of the wrong canonical pages,
  // obtaining the correct effect.
  if ((address_space->get_type() == AS_TYPE_FILE) &&
      (address_space->get_size() > kernel_record->file_size)) {

    fprintf(stderr,
	    "WARNING: Performing unrecorded truncation/deletion\n");
    handle_file_truncate();

  }
	
  // If the trace record's length is longer than ours (e.g., for files
  // we've not seen before), adopt the new, longer size.
  address_space->expand_size(kernel_record->file_size);

} // Consumer::handle_file_open
// ===================================================================



// ===================================================================
void
Consumer::handle_file_read_or_write () {
	
  // Get the address space for the file.
  Address_Space* address_space =
    safe_file_ID_to_AS(kernel_record->file_ID, FILE_TYPE_UNKNOWN);

  // If the address space is uncached, there's nothing to do here.
  if (address_space->get_type() == AS_TYPE_UNCACHED) {
    return;
  }
	
  // Traverse each page in the requested read or write operation.
  virtual_page_ID_t page = kernel_record->starting_file_page;
  while (page < kernel_record->ending_file_page) {

    // Has a canonical page yet been assigned to this page?
    Canonical_Page* canonical_page =
      address_space->lookup_canonical_page(page);
    if (canonical_page == NULL) {
			
      // No.  Is this an access that elongates a real file?
      if ((address_space->get_type() == AS_TYPE_FILE) &&
	  (page >= address_space->get_size())) {
				
	// Increase the file size accordingly.
	address_space->expand_size(page + 1);
				
	// Change the tag to indicate that this was an elongating read
	// or write.  Note that an `elongating read' is just the first
	// opportunity to observe an access to a file that has been
	// elogated (perhaps with a truncate() call that actually
	// elongates).
	kernel_record->tag = (kernel_record->tag == TAG_FILE_READ ?
			      TAG_FILE_ELONGATING_READ :
			      TAG_FILE_ELONGATING_WRITE);
				
      }
			
      // Create a new canonical page and map it into the file's space.
      canonical_page = create_canonical_page(kernel_record->file_ID,
					     page);
      address_space->insert_mapping(page, canonical_page);
			
    } else {
			
      // Yes, the canonical page exists.  Sanity check: The page
      // should be a file page.  Furthermore, this page must fall
      // within the known size of the address space, and it must
      // correctly belong to this address space.
      assert(canonical_page->get_type() == PAGE_TYPE_FILE);
      assert(page < address_space->get_size());
      assert(canonical_page->get_file_ID() == kernel_record->file_ID);
      assert(canonical_page->get_virtual_page_ID() == page);

    }

    // Advance to the next page of the file.
    page++;

  }

} // Consumer::handle_file_read_or_write
// ===================================================================



// ===================================================================
void
Consumer::handle_file_delete () {
	
  // Attempt to grab the address space for this file.
  file_ID_to_AS_map_t::iterator i =
    file_ID_to_AS_map.find(kernel_record->file_ID);
	
  // It may be the case that we've not encountered this file before,
  // as a file need not be mapped or opened to be deleted by some
  // process.  There's only work to do if we've a record of this
  // file.
  if (i != file_ID_to_AS_map.end()) {
		
    Address_Space* address_space = (*i).second;
		
    // Delete the address space and its association with the given
    // file ID.
    delete address_space;
    address_space = NULL;
    file_ID_to_AS_map.erase(i);
		
  }
	
} // Consumer::handle_file_delete
// ===================================================================



// ===================================================================
void
Consumer::handle_file_truncate () {

  // Attempt to find the address space for this file.  There is only
  // work to do if an address space exists.
  file_ID_to_AS_map_t::iterator i =
    file_ID_to_AS_map.find(kernel_record->file_ID);
  if (i != file_ID_to_AS_map.end()) {

    // Grab the address space for the file and the new file length.
    Address_Space* address_space = (*i).second;
    file_offset_t new_size = kernel_record->file_size;

    // Sanity check: We cannot truncate to a longer length.
    assert(new_size <= address_space->get_size());

    // Walk the pages from the new length to the old, deleting each.
    for (file_offset_t current_page = new_size;
	 current_page < address_space->get_size();
	 current_page++) {

      // Sanity check: If this virtual page maps to a real canonical
      // page, then that page should have only this one mapping left.
      Canonical_Page* canonical_page =
	address_space->lookup_canonical_page(current_page);
      if ((canonical_page != NULL) &&
	  (canonical_page->get_mapping_count() != 1)) {
	fprintf(stderr,
		"WARNING: Consumer::handle_file_truncate(): ");
	fprintf(stderr,
		"Making zombie file page with mapping count %d != 1\n",
		canonical_page->get_mapping_count());
	canonical_page->make_zombie();
      }

      // Delete the mapping and the canonical page (if it existed).
      address_space->remove_mapping(current_page);

    }
	
    // Adopt the new length.
    address_space->contract_size(new_size);

  }

} // Consumer::handle_file_truncate
// ===================================================================



// ===================================================================
void
Consumer::handle_invalid () {
	
  fprintf(stderr, "Consumer::handle_invalid(): ");
  fprintf(stderr,
	  "Invalid kernel event tag %c(%d) at record %qd\n",
	  kernel_record->tag,
	  (int)kernel_record->tag,
	  kernel_record->record_number);
  exit(1);
	
} // Consumer::handle_invalid
// ===================================================================



// ===================================================================
void
Consumer::update_with_reference_record () {
	
  // If this is the first observation of a reference, mark the change
  // of phases to waiting for a scheduling event.
  if (merge_phase == PHASE_PRE_REFERENCE) {
    merge_phase = PHASE_WAITING_FOR_SCHEDULE;
  }
	
  // If we're still waiting to see the first scheduling event (which
  // would tell us which task is active), then we cannot yet process
  // this record.
  if (merge_phase == PHASE_WAITING_FOR_SCHEDULE) {
    return;
  }
	
  // Grab local copies of the referenced page's information.
  const context_ID_t context_ID =
    reference_record->virtual_page.context_ID;
  const virtual_page_ID_t page_ID =
    reference_record->virtual_page.page_ID;
	
  // Sanity check: The address space used should be the scheduled
  // task's.
  if (context_ID != scheduled_task->process->context_ID) {
    fprintf(stderr, "Consumer::update_with_reference_record(): ");
    fprintf(stderr,
	    "Unexpected context ID used at record %qd\n",
	    reference_record->record_number);
    exit(1);
  }
	
  // Sanity check: The referenced page cannot be in the kernel's
  // space.
  assert(page_ID < KERNEL_VIRTUAL_PAGE_OFFSET);
	
  // Grab the map of the context in which this reference was
  // performed.
  Process* process = safe_context_ID_to_process(context_ID);
  Address_Space* address_space = process->address_space;
	
  // Sanity check: The reference must be to some mapped page.  Since
  // all mappings are recorded in the kernel trace including brk()
  // calls, no references should exist to unmapped pages.
  Canonical_Page* canonical_page =
    address_space->safe_lookup_canonical_page(page_ID);
	
  // If this is a write operation to the zero page, then we must
  // create a new canonical page.  Note that in the real OS, the page
  // must initially not have ever been mapped to the zero page -- we
  // just assumed it as a default for the anonymous pages.  This
  // operation is effectively performing a COW from a canonical page
  // that the original system never mapped.
  if (((reference_record->tag == TAG_SINGLE_WRITE) ||
       (reference_record->tag == TAG_MULTIPLE_WRITE)) &&
      (canonical_page == &canonical_zero_page)) {
		
    // Unmap the canonical page.
    address_space->remove_mapping(page_ID);
		
    // Create a new anonymous, canonical page and map to it.
    canonical_page = create_canonical_page(PAGE_TYPE_ANONYMOUS);
    address_space->insert_mapping(page_ID, canonical_page);
		
  }
	
} // Consumer::update_with_reference_record
// ===================================================================



// ===================================================================
void
Consumer::update_with_kernel_record () {
	
  handler_t handler =
    handler_dispatch_table[(int)kernel_record->tag];
  if (handler == (handler_t)NULL) {
		
    fprintf(stderr, "Consumer::update_with_kernel_record(): ");
    fprintf(stderr,
	    "At record %qd: Unknown tag %c(%d)\n",
	    kernel_record->record_number,
	    kernel_record->tag,
	    (int)kernel_record->tag);
    exit(1);
		
  }
	
  ((*this).*(handler))();
	
} // Consumer::update_with_kernel_record
// ===================================================================



// ===================================================================
void
Consumer::initialize () {
	
  // Create a task object for the initial kernel task as well as a
  // process object for the ``kernel process''.
  kernel_task = allocate_new_task();
  Process* kernel_process = allocate_new_process();
	
  // Give each an ID of 0, initialize the other fields, and assign the
  // thread to the process.
  kernel_task->task_ID = 0;
  kernel_task->process = kernel_process;
  kernel_task->parent_task = NULL;
  kernel_process->process_ID = KERNEL_PROCESS_ID;
  kernel_process->task_list.push_front(kernel_task);
  kernel_process->image_file_ID.inode_ID = 0;
  kernel_process->image_file_ID.major = 0;
  kernel_process->image_file_ID.minor = 0;
  strcpy(kernel_process->image_pathname, "***KERNEL***");
	
  // Kernel threads may create mappings into the kernel's virtual
  // address space.  We won't see any references to those, but we'll
  // process them mappings.
  kernel_process->context_ID = KERNEL_CONTEXT;
  kernel_process->address_space = new Address_Space(AS_TYPE_VIRTUAL);
  assert(kernel_process->address_space != NULL);
	
  // Insert into both the task_ID- and context_ID-keyed tables.
  task_ID_to_Task_map[kernel_task->task_ID] = kernel_task;
  context_ID_to_Process_map[kernel_process->context_ID] =
    kernel_process;
	
  // Assume that the base kernel task is the first one scheduled.
  scheduled_task = kernel_task;
	
}
// ===================================================================



// ===================================================================
Task*
Consumer::allocate_new_task () {
	
  // Attempt to allocate a new Task.  Check for success.
  Task* task = new Task;
  if (task == NULL) {
    fprintf(stderr, "Consumer::allocate_new_task(): ");
    fprintf(stderr, "Allocation failure\n");
    exit(1);
  }
	
  return task;
	
} // Consumer::allocate_new_task
// ===================================================================



// ===================================================================
Process*
Consumer::allocate_new_process () {
	
  // Attempt to allocate a new Task.  Check for success.
  Process* process = new Process;
  if (process == NULL) {
    fprintf(stderr, "Consumer::allocate_new_process(): ");
    fprintf(stderr, "Allocation failure\n");
    exit(1);
  }
	
  return process;
	
} // Consumer::allocate_new_process
// ===================================================================
