/**
 * \file Consumer.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date June 2004
 * \brief Header of base class for reading traces and rebuilding
 *        system state.
 *
 * A class used to read both a kernel and reference trace and, with
 * each record, update data structures that reflect the state of the
 * tasks and the memory.
 *
 * Note that this is an ABSTRACT CLASS: A subclass, by defining the
 * pure virtual methods found in this class, can choose to maintain
 * extra data structures or to produce output before processing, at
 * the processing of each record, and after processing.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_CONSUMER_HH)
#define _CONSUMER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <stdio.h>
#include "Address_Space.hh"
#include "Process.hh"
#include "Reader.hh"
#include "Task.hh"
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * \brief Process the reference and kernel traces.
 *
 * The main program creates one of these objects which is then
 * designed to read, parse, and process the two input traces.  It
 * contains all of the data structures that define the state of the
 * system at each moment.  It is an abstract class that must be
 * extended in order to act on the state after each record is
 * processed.
 **/
class Consumer {

public: // CONSTRUCTOR & DESTRUCTOR

  /**
   * \brief The constructor.
   * \param reader The object that will read and parse the
   *        reference and kernel traces.
   *
   * Set up a Consumer object to process a specific reference/kernel
   * trace pair.
   **/
  Consumer (Reader* const reader);

  /**
   * \brief The deconstructor.
   *
   * Deconstruct a Consumer.  Close the input files.
   **/
  virtual ~Consumer ();

public: // METHODS

  /**
   * \brief Merge the reference and kernel traces.
   *
   * Perform the merging.  This is the entry point to the real
   * computational work.
   **/
  void merge ();

protected: // METHODS

  /**
   * \brief Associate a particular tag with a handler method.
   * \param tag The tag that will trigger a call of the given handler.
   * \param handler The handler method to be called when the tag is
   *         encountered.
   * \see handler_dispatch_table
   **/
  void assign_handler (tag_t tag, handler_t handler);

  /**
   * \brief Lookup a task object via its ID.
   * \param task_ID The ID of the task used for lookup.
   * \return A pointer to the task object with the given ID.
   *
   * Lookup a task object given a task ID.  If the lookup fails, then
   * abort processing.
   **/
  Task* safe_task_ID_to_task (const task_ID_t task_ID);

  /**
   * \brief Lookup a process via its context ID.
   * \param context_ID The context ID for which to search.
   * \return A pointer to the process using the given context ID.
   *
   * Lookup a process object given a context ID.  If no process is
   * associated with the context, then processing is aborted.
   **/
  Process* safe_context_ID_to_process (const context_ID_t context_ID);

  /**
   * \brief Determine whether a task is live during synchronization.
   * \see synchronize()
   * \param task_ID The ID of the task to examine.
   * \return Whether the task is live during syncrhonization.
   *
   * When the reference and kernel trace are synchronized, some tasks
   * that were created prior to the first reference may still be live,
   * and others may have terminated.  During synchronization, a table
   * of live and dead processes is kept.  This method returns the
   * status of a task as synchronization is being performed.  If the
   * given task has never been created, processing is aborted.
   **/
  bool safe_live (const task_ID_t task_ID);

  /**
   * \brief Lookup the address space for a file.
   * \param file_ID The file to lookup.
   * \param file_type The type of the address space.  This parameter
   *        should take on a meaningful value if the space may need to
   *        be created (e.g. an open() or mmap() operation), and
   *        FILE_TYPE_UNKNOWN otherwise.
   * \return A pointer to the address space.
   *
   * This method either finds and returns the address space or it
   * aborts processing.  Thus, there is no need to test the return
   * value.
   **/
  Address_Space* safe_file_ID_to_AS (const file_ID_s& file_ID,
				     const file_type_t file_type);

  /**
   * \brief Lookup the address space for an IPC shm segment.
   * \param shm_ID The IPC shared memory segment to lookup.
   * \return A pointer to the address space.
   *
   * This method either finds and returns the address space or it
   * aborts processing.  Thus, there is no need to test the return
   * value.
   **/
  Address_Space* safe_shm_ID_to_AS (const shm_ID_t& shm_ID);

  /**
   * \brief Delete a given task.
   * \param task A pointer to the task to be deleted.
   **/
  void delete_task (Task* task);

  /**
   * \brief Delete a process and its context if they are defunct.
   * \param process_handle The process to examine and possibly delete.
   *
   * If the given process has no tasks assigned to it, ensure that it
   * has no mapped pages, delete its context, its context_ID->process
   * mapping, and the process object itself.  Note that a handle is
   * passed so that the caller's pointer can be cleared if the process
   * is deleted.
   **/
  void delete_if_defunct_process (Process** const process_handle);

  /**
   * \brief Delete any IPC shared memory segments marked as defunct.
   *
   * Called after each kernel record is processed in the event that
   * unmapping occurred that caused an IPC shared memory segment to
   * become defunct.  Those defunct segments are then deleted by this
   * method which eliminates both the underlying canonical pages and
   * the address space.
   **/
  void delete_defunct_shm_segments ();

  /**
   * \brief If defunct, mark an IPC shm segment for deletion.
   * \param canonical_page A canonical page from the segment in
   *        question.
   * \see handle_munmap_range()
   * \see delete_defunct_shm_segments()
   *
   * Examine the given canonical page that is part of an IPC shared
   * memory segment.  If it has a reference count of zero, then the
   * segment is defunct and should be marked for deletion.  This
   * method is called as pages are unmapped, as it is the unmapping
   * that can make an IPC shm segment defunct.  After unmapping is
   * complete, this function is called to determine if there are any
   * defunct segments to be deleted.
   **/
  void mark_if_defunct_shm_segment
  (const Canonical_Page* const canonical_page);

  /**
   * \brief Unlink a task and its process.
   * \param task The task to be unlinked from its process.
   *
   * Given a task and its process, remove that task from the process's
   * task list and nullify that tasks's pointer to the process.
   * Delete the process and its context if it becomes defunct as a
   * result.
   **/
  void unlink_task_and_process (Task* const task);

  /**
   * \brief The default, no-op handler function.
   * 
   * For those kernel events for which no action must be taken, this
   * method, which does nothing, is placed in the handler dispatch
   * table.
   **/
  void handle_no_op () {}

  /**
   * \brief Process the scheduling of a task on the CPU.
   * \see scheduled_task
   *
   * Update the scheduled task based on a kernel event that indicates
   * that a new user-level task was scheduled.  Note that we do
   * <b>not</b> update the active task, as only a boundary crossing
   * event can do that.
   **/
  void handle_schedule ();

  /**
   * \brief Process the creation of a new task.
   *
   * Simply place an entry in the task_ID->process map indicating that
   * this task exists.  Do not associate it with a specific process
   * until it is assigned a context.
   **/
  void handle_fork ();

  /**
   * \brief Process the execution of an application within a task.
   *
   * Assign the process for the given task a new executable image.
   * Note that a complete unmapping event preceeds each exec(), as the
   * exec() system call requests it.  Thus this method does not need
   * to clear the task's existing address space
   **/
  void handle_exec ();

  /**
   * \brief Process the termination of a task.
   * \see delete_if_unused_context()
   * \see handle_complete_unmap()
   *
   * Delete the given task and, if it was the only task of a process,
   * delete both the process and its context.
   **/
  void handle_exit ();

  /**
   * \brief Assign a new context to a task.
   *
   * Assign an existing task to a new context.  Furthermore, use that
   * context to identify the process for which this task is a thread,
   * and associate the task with the process.  If the context being
   * assigned is a new one (unused by any other task), then create the
   * tables that represent that context.  If the old context becomes
   * defunct (that is, unused by any task) as a result, delete the old
   * context.
   **/
  void handle_context_assignment ();

  /**
   * \brief Duplicate the mappings within an address range from one
   *        context to the other.
   *
   * Given source and destination contexts, duplicate the
   * virtual->canonical mappings for a range of pages.  Update the
   * canonical pages to reflect the sharing.  Duplications are the
   * result of the copy-on-write mechanism used during the fork()
   * system call.
   **/
  void handle_duplicate_range ();

  /**
   * \brief Process the mapping of a range of file pages into a range
   *        of virtual pages.
   * \param task_ID The task performing the mapping.
   * \param file_ID The file being mapped.
   * \param starting_virtual_page The virtual page at which the
   *        mapping begins.
   * \param ending_virtual_page The virtual page at which the mapping
   *        ends (non-inclusive)
   * \param starting_file_page The starting page within the file to be
   *         mapped
   * \param ending_file_page The ending page within the file to be
   *        mapped
   * \param file_type The type of the file being mapped.
   * \see handle_file_mmap_range()
   * \see handle_file_delete()
   *
   * Reconstruct the mapping of a sequence of pages from a file into
   * the virtual address space of a task (that is, its context).  This
   * method first ensures that there are canonical pages associated
   * with each file page to be mapped, and that the file has its own
   * address space that assocaite the file pages to the canonical
   * pages.  It then maps the task's virtual pages to those same
   * canonical pages.
   **/
  void do_file_mmap_range (const task_ID_t task_ID,
			   const file_ID_s& file_ID,
			   const virtual_page_ID_t starting_virtual_page,
			   const virtual_page_ID_t ending_virtual_page,
			   const virtual_page_ID_t starting_file_page,
			   const virtual_page_ID_t ending_file_page,
			   const file_type_t file_type);

  /**
   * \brief Update a virtual->file mapping based on a kernel event.
   * \see do_file_mmap_range()
   *
   * Call on do_file_mmap_range() to map a range of virtual pages onto
   * a range of file pages.
   **/
  void handle_file_mmap_range ();

  /**
   * \brief Process the mapping of an anonymous range of virtual
   *        pages.
   * \param task_ID The task performing the mapping.
   * \param starting_virtual_page The virtual page at which the
   *        mapping begins.
   * \param ending_virtual_page The virtual page at which the mapping
   *        ends (non-inclusive)
   * \see handle_anonymous_mmap_range()
   * \see TYPE_ZERO_PAGE
   *
   * It is possible to perform an mmap() (or to call for a brk(),
   * etc.) without a file specified.  In this case, the task's range
   * of virtual pages is mapped to the zero page in copy-on-write
   * fashion.
   **/
  void do_anonymous_mmap_range
  (const task_ID_t task_ID,
   const virtual_page_ID_t starting_virtual_page,
   const virtual_page_ID_t ending_virtual_page);

  /**
   * \brief Update a virtual, anonymous mapping based on a kernel
   *        event.
   * \see do_anonymous_mmap_range()
   *
   * Call on do_anonymous_mmap_range() to map a range of virtual pages
   * onto a range of new canonical pages.
   **/
  void handle_anonymous_mmap_range ();

  /**
   * \brief Map an IPC shared memory segment into a task's context.
   * \param task_ID The task performing the mapping.
   * \param shm_ID The ID of the shm segment being mapped.
   * \param starting_virtual_page The virtual page at which the
   *        mapping begins.
   * \param ending_virtual_page The virtual page at which the mapping
   *        ends (non-inclusive)
   * \see handle_shmat
   *
   * Map the pages of an IPC shared memory segment into the virtual
   * address space of the calling task.  First ensure that the shm
   * segment has its own address space and that it maps the segment's
   * pages to underlying canonical pages.  Then establish mappings
   * from the task's virtual pages to those same canonical pages.
   **/
  void do_shm_mmap_range (const task_ID_t task_ID,
			  const shm_ID_t shm_ID,
			  const virtual_page_ID_t starting_virtual_page,
			  const virtual_page_ID_t ending_virtual_page);

  /**
   * \brief Update a virtual, anonymous mapping based on a kernel
   *        event.
   * \see do_shm_mmap_range()
   *
   * Call on do_shm_mmap_range() to map a range of virtual pages
   * onto a IPC shared memory segment.
   **/
  void handle_shmat ();

  /**
   * \brief Delete an IPC shared memory segment's canonical pages and
   *        map in response to the segment's removal.
   *
   * To reflect that a shm segment has been deleted, its address space
   * and its underlying canonical pages are deleted.  This method
   * sanity checks that the canonical pages are defunct and aborts
   * processing if one or more is not.
   **/
  void handle_shm_destroy ();

  /**
   * \brief Update a file's address space based on its being opened.
   *
   * Upon opening, grab a file's address space (or create it if
   * necessary) and update our record of its length as reflected by
   * the opening operation.
   **/
  void handle_file_open ();

  /**
   * \brief Remap a virtual page range to another location.
   *
   * For a given task and its context, move a range of virtual pages.
   * Specifically, unmap the original range and create a new mapping
   * within the new range, all the while maintaining the mapping to
   * the underlying canonical pages used in the old mapping.  Note
   * that the new range may be larger than the original, and thus new
   * pages will require the creation of new canonical space.
   **/
  void handle_mremap ();

  /**
   * \brief Unmap a range of virtual pages.
   * \see delete_if_unused_page()
   * \see mark_if_defunct_shm_segment()
   *
   * For a given task and its context, unmap a range of virtual pages
   * by deleting their mapping to canonical pages.  It is not an
   * error to unmap a page that has no mapping.  If the unmapping of
   * some anonymous, canonical page causes that page to be defunct,
   * it may be deleted.  If an IPC shared memory segment is unmapped
   * and thus becomes defunct, that segment is marked for deletion.
   **/
  void handle_munmap_range ();

  /**
   * \brief Unmap every virtual page in the context specified in a
   *        kernel event.
   *
   * Given the context ID in the kernel event, look up the address
	 * space and delete it.
   **/
  void handle_complete_unmap ();

  /**
   * \brief Split a COW page that has been written.
   *
   * For a page that has been mapped COW, delete its previous
   * mapping, create a new canonical page to represent the new copy
   * created, and establish a virtual->canonical mapping to that new
   * page at the same virtual address.
   **/
  void handle_cow_unmap ();

  /**
   * \brief Ensure that the read/written files pages have
   *        corresponding canonical page IDs.
   **/
  void handle_file_read_or_write ();

  /**
   * \brief Delete a file's canonical pages and map in response to
   *        the file's removal.
   *
   * To reflect that a file has been deleted, its address space and
   * its underlying canonical pages are deleted.  This method sanity
   * checks that the canonical pages are defunct and aborts processing
   * if one or more is not.
   **/
  void handle_file_delete ();

  /**
   * \brief Truncate a file by discarding some of its trailing pages.
   *
   * To reflect that a file has been truncated, some number of its
   * trailing canonical pages are unmapped and deleted, and its size
   * shortened.  This method checks that the canonical pages are not
   * otherwise mapped before deleting them, and will abort processing
   * otherwise.
   **/
  void handle_file_truncate ();

  /**
   * \brief Handle an invalid reference tag.
   *
   * For any kernel event record that contains an invalid tag, emit
   * the error and abort.
   **/
  void handle_invalid ();

  /**
   * \brief Update the state based on the current reference record.
   * \see reference_record
   *
   * This method is called after each reference record is read so that
   * the state of the system can be updated in response to the
   * reference.
   **/
  void update_with_reference_record ();

  /**
   * \brief Update the state based on the current kernel record.
   * \see kernel_record
   *
   * This method is called after each kernel record is read so that
   * the state of the system can be updated in response to the event.
   **/
  void update_with_kernel_record ();

protected: // VIRTUAL METHODS

  /**
   * \brief Called before processing begins.
   *
   * Performs remaining initialization.  In the base class, this
   * creates a kernel task and process, and inserts those into the
   * appropriate maps.  These steps cannot be performed in the
   * constructor because they rely on calls to virtual methods that
   * cannot be invoked from the constructor itself (because the
   * subclass constructors have not completed).
   **/
  virtual void initialize ();

  /**
   * \brief Allocate a new task object.
   * \return A pointer to a newly-allocated task object.
   *
   * Allocate a new task object, where the object can be a Task or one
   * of its subclasses.  Abort if the allocation fails.
   **/
  virtual Task* allocate_new_task ();

  /**
   * \brief Allocate a new process object.
   * \return A pointer to a newly-allocated process object.
   *
   * Allocate a new process object, where the object can be a Process
   * or one of its subclasses.  Abort if the allocation fails.
   **/
  virtual Process* allocate_new_process ();

protected: // PURE VIRTUAL METHODS

  /**
   * \brief Called after processing ends.
   *
   * A pure virtual method that is invoked after processing of both
   * traces is complete.  A subclass must implement this method and,
   * if desired, use it to clean up and complete its processing.
   **/
  virtual void clean_up () = 0;

  /**
   * \brief Called after processing each reference record.
   * \see reference_record
   * \see Reader
   * \see update_maps_with_reference_record()
   *
   * A pure virtual method that is invoked after each reference
   * record is read and processed (that is, used to update the state
   * of system).
   **/
  virtual void act_on_reference_record () = 0;

  /**
   * \brief Called after processing each kernel record.
   * \see kernel_record
   * \see read_kernel_record()
   * \see update_maps_with_kernel_record()
   *
   * A pure virtual method that is invoked after each kernel record
   * is read and processed (that is, used to update the state of
   * system.
   **/
  virtual void act_on_kernel_record () = 0;

protected: // DATA

  /** \brief The Reader used to input the reference sequence. */
  Reader* reader;

  /** \brief The base kernel task. */
  Task* kernel_task;

  /**
   * \brief The most recent record from the reference trace.
   * \see Reader
   *
   * A single structure into which each reference record is read.
   * The values are overwritten when each new record is parsed.
   **/ 
  reference_record_s* reference_record;

  /**
   * \brief The current phase of merging.
   **/
  merge_phase_t merge_phase;

  /**
   * \brief The most recent record from the kernel trace.
   * \see read_kernel_record()
   *
   * A single structure into which each kernel record is read.  The
   * values are overwritten when each new record is parsed.
   **/ 
  kernel_record_s* kernel_record;

  /**
   * \brief The task currently scheduled by the kernel's
   * scheduler.
   **/
  Task* scheduled_task;

  /**
   * \brief The next available process ID.
   *
   * Note that these IDs are not taken from the task IDs that exist in
   * the kernel trace.  These IDs are assigned to processes linearly
   * as each process is discovered.
   **/
  process_ID_t next_process_ID;

  /** \brief A table that maps task IDs to task objects. */
  task_ID_to_Task_map_t task_ID_to_Task_map;

  /** \brief A table that maps context IDs to task objects. */
  context_ID_to_Process_map_t context_ID_to_Process_map;

  /**
   * \brief A table used to track the liveness of tasks.
   * \see safe_live()
   * \see synchronize()
   *
   * A per-task table of the liveness of tasks that is maintained
   * during synchronization of the input traces.
   **/
  live_map_t live_map;

  /** \brief A table used to map file IDs to address spaces. */
  file_ID_to_AS_map_t file_ID_to_AS_map;

  /** \brief A table used to map IPC shared memory segment IDs to
   *         address spaces. */
  shm_ID_to_AS_map_t shm_ID_to_AS_map;

  /** \brief A dispatch table that maps tags to handler functions. */
  handler_t* handler_dispatch_table;

  /** \brief A buffer used for input and parsing. */
  char* buffer;

}; // class Consumer
// ===================================================================



// ===================================================================
#endif // _CONSUMER_HH
// ===================================================================
