/**
 * \file Per_Process_Consumer.cc
 * \author Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
 * \date September 2004
 * \brief Produces per-process reference files.
 **/



// ===================================================================
// INCLUDES

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "Per_Process_Consumer.hh"
#include "Canonical_Page.hh"
// ===================================================================



// ===================================================================
Per_Process_Consumer::Per_Process_Consumer
(Reader* const reader,
 const char* const output_directory_pathname) : Consumer(reader) {

   // Allocate space for the output directory pathname.
   this->output_directory_pathname = new char[BUFFER_SIZE];
   if (output_directory_pathname == NULL) {
     fprintf(stderr, "Per_Process_Consumer::Per_Process_Consumer():");
     fprintf(stderr, "Could not allocate space for output directory\n");
     exit(1);
  }

  // Copy the output directory pathname into this object.
  strcpy(this->output_directory_pathname, output_directory_pathname);

} // Per_Process_Consumer::Per_Process_Consumer
// ===================================================================



// ===================================================================
Per_Process_Consumer::~Per_Process_Consumer () {

  // Traverse the task_ID->Task table.  Note that we have no
  // process_ID->Process table, so we must find the processes via
  // their tasks.
  task_ID_to_Task_map_t::iterator i = task_ID_to_Task_map.begin();
  while (i != task_ID_to_Task_map.end()) {

    // If it exists, delete the per-process writer associated with
    // this task's process, thus flushing and closing its output.
    Task* task = (*i).second;
    Per_Process_Process* process = (Per_Process_Process*)task->process;
    if (process->writer != NULL) {
      // delete process->writer;
      // process->writer = NULL;
    }

    // Move to the next task.
    i++;

  } // task_ID->Task table traversal

} // Per_Process_Consumer::~Per_Process_Consumer
// ===================================================================



// ===================================================================
void
Per_Process_Consumer::act_on_reference_record () {

  // If we're still waiting to see the first scheduling event (which
  // would tell us which task is active), then we cannot yet process
  // this record.
  assert(merge_phase != PHASE_PRE_REFERENCE);
  if (merge_phase != PHASE_PROCESSING_REFERENCES) {
    return;
  }

  // Lookup the canonical page ID for this reference in the scheduled
  // task's context.
  Address_Space* address_space = scheduled_task->process->address_space;
  virtual_page_ID_t virtual_page_ID =
    reference_record->virtual_page.page_ID;
  Canonical_Page* canonical_page =
    address_space->safe_lookup_canonical_page(virtual_page_ID);
  canonical_page_ID_t canonical_page_ID = canonical_page->get_ID();

  // Submit this reference to this process's writer only for
  // non-uncacheable pages.
  if (canonical_page != &canonical_uncacheable_page) {

    Per_Process_Process* process =
      (Per_Process_Process*)scheduled_task->process;
    if (process->writer == NULL) {
      create_writer(process);
    }
    process->writer->
      append_reference_record(reference_record->tag,
			      canonical_page_ID,
			      process->virtual_cycle_time,
			      process->virtual_reference_time);

  }

} // Consumer::act_on_reference_record
// ===================================================================



// ===================================================================
void
Per_Process_Consumer::act_on_kernel_record () {

  // Determine how to react to this record depending on the type of
  // kernel event.
  switch (kernel_record->tag) {

  case TAG_FILE_READ:
  case TAG_FILE_WRITE: {

    // These kernel events are really referencing events, so let's
    // treat them as such.  Lookup the address space for the file.
    Address_Space* address_space =
      safe_file_ID_to_AS(kernel_record->file_ID, FILE_TYPE_UNKNOWN);

    // Traverse each page in the requested read or write operation.
    virtual_page_ID_t page = kernel_record->starting_file_page;
    while (page < kernel_record->ending_file_page) {

      // Lookup the canonical page underlying this file page.
      Canonical_Page* canonical_page =
	address_space->safe_lookup_canonical_page(page);
      canonical_page_ID_t canonical_page_ID = canonical_page->get_ID();

      // Submit this reference to this process's writer only for
      // non-uncacheable pages.
      if (canonical_page != &canonical_uncacheable_page) {

	Per_Process_Process* process =
	  (Per_Process_Process*)scheduled_task->process;
	if (process->writer == NULL) {
	  create_writer(process);
	}
	process->writer->
	  append_reference_record(kernel_record->tag,
				  canonical_page_ID,
				  process->virtual_cycle_time,
				  process->virtual_reference_time);

      }

      // Advance to the next page of the file.
      page++;

    }

    break;

  }

  default:

    // Do nothing -- ignore this kernel record, as it's not relevant
    // to the output of these per-task traces.
    break;

  }

} // Consumer::act_on_kernel_record
// ===================================================================



// ===================================================================
Process*
Per_Process_Consumer::allocate_new_process () {

  // Attempt to allocate a Per_Process_Process, a specific subclass
  // designed for use with this class.
  Process* process = new Per_Process_Process;
  if (process == NULL) {
    fprintf(stderr, "Per_Process_Consumer::allocate_new_process():");
    fprintf(stderr, "Allocation failure\n");
    exit(1);
  }

  return process;

} // Per_Process_Consumer::allocate_new_process
// ===================================================================



// ===================================================================
void
Per_Process_Consumer::create_writer (Per_Process_Process* process) {

  // Create a new writer for this process.
  process->writer = new Per_Process_Writer(output_directory_pathname,
					   process->process_ID);
  if (process->writer == NULL) {
    fprintf(stderr, "Per_Process_Consumer::create_writer(): ");
    fprintf(stderr, "Failed to allocate new per-process writer\n");
    exit(1);
  }

} // Per_Process_Consumer::create_writer
// ===================================================================
