/**
 * \file Task.hh
 * \author Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
 * \date June 2004
 * \brief A structure to store data about a task.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_TASK_HH)
#define _TASK_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
// FORWARD DECLARATION

class Process;
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * \brief A structure to store data about a task.
 * \see Process
 *
 * Tasks are the Linux equivalent of a kernel-level thread.  One or
 * more tasks, using a virtual address space, constitute a
 * process. Note that we make data members public for simplicity of
 * access, violating good OO programming guidelines.
 **/
class Task {

public: // CONSTRUCTOR & DESTRUCTOR (if any)

  /**
   * \brief The constructor.
   *
   * Initialize a task by zeroing its fields.
   **/
  Task ();

  /**
   * \brief The destructor.
   *
   * Nothing to do here.
   **/
  virtual ~Task () {}

  /**
   * \brief Record the cycle time at the start of a quanta.
   * \param system_cycle_time The system-wide CPU cycle time at which
   *        the quanta starts.
   *
   * Help the task to maintain its virtual cycle time based on the
   * provided global cycle time.  Note that reference time is not
   * considered because it is provided as a virtual value (i.e. per
   * task), so it cannot have changed since the scheduled task's last
   * scheduling.
   **/
  void mark_quanta_start (const timestamp_t system_cycle_time);

  /**
   * \brief Record time that has passed for this task.
   * \param system_cycle_time The CPU cycle time
   * \param virtual_reference_time The reference time of the current
   *        task.
   *
   * Update the system CPU time and the virtual CPU and reference
   * times according to the time that has passed since the last
   * observed trace record.
   **/
  void mark_time (const timestamp_t system_cycle_time,
		  const timestamp_t virtual_reference_time);

public: // DATA MEMBERS

  /** \brief A unique task number. */
  task_ID_t task_ID;

 /** \brief A pointer to the process to which this task belongs. */
  Process* process;

 /** \brief The parent from which this task was forked. */
  Task* parent_task;

  /** \brief The cycle time that has passed for this task alone. */
  timestamp_t virtual_cycle_time;

  /** \brief The reference time that has passed for this task
   *         alone. */
  timestamp_t virtual_reference_time;

  /** \brief The most recently observed system cycle time. */
  timestamp_t last_system_cycle_time;

}; // class Task
// ===================================================================



// ===================================================================
#endif // _TASK_HH
// ===================================================================
