/**
 * \file V0_4_Reader.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date June 2004
 * \brief Code for a class that reads kVMTrace v0.4 traces.
 *
 * A subclass of Reader that can read the reference and kernel traces
 * produced by kVMTrace v0.4.  See the header file for more details.
 */



// ===================================================================
// INCLUDES

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "V0_4_Reader.hh"
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
V0_4_Reader::V0_4_Reader (const char* const reference_trace_pathname,
			  const char* const kernel_trace_pathname) :
  Reader(reference_trace_pathname,
	 kernel_trace_pathname) {

  // Initialize the previous values (some fields of v0.4 reference
  // traces are difference encoded) to 0.
  previous_cycle_time = 0;
  previous_reference_time = 0;
  previous_context_ID = 0;

} // V0_4_Reader
// ===================================================================



// ===================================================================
void
V0_4_Reader::read_reference_record
(reference_record_s* const reference_record) {

  reference_record_number++;

  // Attempt to read the next record of the reference trace.
  char* read_result = fgets(buffer, BUFFER_SIZE, reference_stream);

  // Did the read fail?
  if (read_result == NULL) {

    // Distinguish between an EOF and other errors.
    if (feof(reference_stream)) {

      // And EOF simply indicates the end of the trace.  Artificially
      // set the tag accordingly.
      reference_record->tag = TAG_END_OF_TRACE;
      return;

    } else {

      // Other failures to read are the sign of a real problem.
      // Abort!
      fprintf(stderr, "V0_4_Reader::read_reference_record(): ");
      fprintf(stderr,
	      "Read error at record %qx\n",
	      reference_record_number);
      exit(1);

    }

  }

  // Parse the record type, which should be the first character of the
  // buffer.
  reference_record->tag = buffer[0];

  // Attempt to parse the rest of the record based on its type.
  // Declare and assign some variables here to avoid compiler
  // complaints of skipping initialization in switch-case statement.
  unsigned int parse_result = 0;
  unsigned int items_to_parse = 0;
  timestamp_t cycle_time_delta = 0;
  timestamp_t reference_time_delta = 0;
  context_ID_t context_ID_delta = 0;
  switch (reference_record->tag) {

  case TAG_SINGLE_READ:
  case TAG_MULTIPLE_READ:
  case TAG_SINGLE_WRITE:
  case TAG_MULTIPLE_WRITE:

    // Read the remainder of the record.
    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %lx %lx",
			  &cycle_time_delta,
			  &reference_time_delta,
			  &context_ID_delta,
			  &reference_record->virtual_page.page_ID);

    // Update the timestamps (difference encoded).
    previous_cycle_time += cycle_time_delta;
    previous_reference_time += reference_time_delta;

    // Update the context ID (change encoded).
    if (context_ID_delta != 0) {
      previous_context_ID = context_ID_delta;
    }

    // Assign values into the record.
    reference_record->cycle_timestamp = previous_cycle_time;
    reference_record->reference_timestamp = previous_reference_time;
    reference_record->virtual_page.context_ID = previous_context_ID;
    reference_record->record_number = reference_record_number;

    break;

  default:

    fprintf(stderr, "V0_4_Reader::read_reference_record(): ");
    fprintf(stderr,
	    "Invalid reference tag %c(%d) at record %qx\n",
	    reference_record->tag,
	    (int)reference_record->tag,
	    reference_record_number);
    exit(1);

  }

  // If the parse failed, emit the error and exit.
  if (parse_result != items_to_parse) {

    fprintf(stderr, "V0_4_Reader::read_reference_record(): ");
    fprintf(stderr,
	    "Could not parse remainder at record %qx\n",
	    reference_record_number);
    exit(1);

  }

} // read_reference_record
// ===================================================================



// ===================================================================
void
V0_4_Reader::read_kernel_record (kernel_record_s* kernel_record) {

  kernel_record_number++;

  // Attempt to read the next record of the kernel trace.
  char* read_result = fgets(buffer, BUFFER_SIZE, kernel_stream);

  // Did the read fail?
  if (read_result == NULL) {

    // Distinguish between an EOF and other errors, but abort
    // processing in either case.  (The kernel trace cannot end before
    // the reference trace, and so its end should never be observed.)
    fprintf(stderr, "V0_4_Reader::read_kernel_record(): ");
    if (feof(kernel_stream)) {
      fprintf(stderr, "Premature EOF for kernel trace\n");
    } else {
      fprintf(stderr, "Read error kernel trace\n");
    }
    exit(1);

  }

  // Parse the record type, which should be the first character of the
  // buffer.
  kernel_record->tag = buffer[0];

  // Attempt to parse the rest of the record based on its type.
  // Declare and assign some variables here to avoid compiler
  // complaints of skipping initialization in switch-case statement.
  unsigned int parse_result = 0;
  unsigned int items_to_parse = 0;
  virtual_address_t starting_virtual_address = 0;
  virtual_address_t ending_virtual_address = 0;
  virtual_address_t old_starting_virtual_address = 0;
  offset_t length = 0;
  offset_t old_length = 0;
  file_offset_t file_offset = 0;
  switch (kernel_record->tag) {

  case TAG_SCHEDULE:
  case TAG_ACCEPT:

    items_to_parse = 3;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID);
    break;

  case TAG_EXIT:

    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->user_execution_time,
			  &kernel_record->system_execution_time);
    break;

  case TAG_FORK:

    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %hx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->parent_task_ID);
    break;

  case TAG_EXEC:

    items_to_parse = 7;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %hx %hx %s",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  kernel_record->filename);
    break;

  case TAG_CONTEXT_ASSIGNMENT:

    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->context_ID);
    break;

  case TAG_DUPLICATE_RANGE:

    items_to_parse = 6;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %lx %lx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->context_ID,
			  &kernel_record->duplicate_context_ID,
			  &starting_virtual_address,
			  &ending_virtual_address);
    break;

  case TAG_MMAP_FILE:

    items_to_parse = 10;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx %lx %hx %hx %qx %s",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  &file_offset,
			  kernel_record->filename);
    break;

  case TAG_MMAP_ANONYMOUS:

    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length);
    break;

  case TAG_MREMAP:

    items_to_parse = 7;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length,
			  &old_starting_virtual_address,
			  &old_length);
    break;

  case TAG_MUNMAP:

    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length);
    break;

  case TAG_COMPLETE_UNMAP:

    items_to_parse = 3;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->context_ID);
    break;

  case TAG_SHMAT:

    items_to_parse = 6;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length,
			  &kernel_record->shm_ID);
    break;

  case TAG_SHM_DESTROY:

    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->shm_ID);
    break;

  case TAG_COW_UNMAP:
  case TAG_DEBUG:
    // Note that debugging was used, in earlier v0.4 versions, to
    // differentiate between COW sites in the kernel.

    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address);
    break;

  case TAG_FILE_OPEN:

    items_to_parse = 7;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %hx %hx %s",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  kernel_record->filename);
    break;

  case TAG_FILE_CLOSE:

    items_to_parse = 6;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %hx %hx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor);
    break;

  case TAG_FILE_READ:
  case TAG_FILE_WRITE:

    items_to_parse = 8;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %hx %hx %qx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  &file_offset,
			  &length);
    break;

  case TAG_FILE_DELETE:

    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %lx %hx %hx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor);
    break;

  case TAG_END_OF_TRACE:

    fprintf(stderr, "V0_4_Reader::read_kernel_record(): ");
    fprintf(stderr, "Premature end of kernel trace\n");
    exit(1);
    break;

  default:
    fprintf(stderr, "V0_4_Reader::read_kernel_record(): ");
    fprintf(stderr,
	    "Invalid kernel tag %c(%d)\n",
	    kernel_record->tag,
	    (int)kernel_record->tag);
    exit(1);

  }

  // If the parse failed, emit the error and exit.
  if (parse_result != items_to_parse) {
    fprintf(stderr, "V0_4_Reader::read_kernel_record(): ");
    fprintf(stderr,
	    "Could not parse remainder at record %qd\n",
	    kernel_record_number);
    exit(1);
  }

  // Many record types specify either a virtual page or a range of
  // virtual pages.  Calculating the one/starting page can be done for
  // every record type, although it may be meaningless for some.  How
  // the ending page is calculated depends on record type, where some
  // provide a length of the range, and others provide an address for
  // the last page in the range.
  kernel_record->starting_virtual_page =
    starting_virtual_address >> PAGE_SIZE_ORDER;  
  switch (kernel_record->tag) {

  case TAG_DUPLICATE_RANGE:

    kernel_record->ending_virtual_page =
      ending_virtual_address >> PAGE_SIZE_ORDER;
    break;

  case TAG_MMAP_FILE:
  case TAG_MMAP_ANONYMOUS:
  case TAG_MREMAP:
  case TAG_SHMAT:
  case TAG_MUNMAP:

    // The length may not be page-aligned, so do so by taking the
    // ceiling w.r.t. the page size.
    length = (length + PAGE_SIZE - 1) & PAGE_NUMBER_MASK;

    // Convert the ending address to a page number.
    kernel_record->ending_virtual_page =
      (starting_virtual_address + length) >> PAGE_SIZE_ORDER;

    // Sanity check: The length cannot overflow the address space.
    assert(kernel_record->starting_virtual_page <=
	   kernel_record->ending_virtual_page);

    break;

  }

  // For some record types, a second offset and length must be
  // converted into starting and ending pages.  For some, the address
  // range is within a file, and for others, it is a second range
  // within a virtual address space.
  switch (kernel_record->tag) {

  case TAG_MMAP_FILE:
    {

    // mmap() records provide the file offset in page numbers.
    unsigned int page_length = length >> PAGE_SIZE_ORDER;
    kernel_record->starting_file_page = file_offset;
    kernel_record->ending_file_page = file_offset + page_length;
    break;

    }

  case TAG_FILE_READ:
  case TAG_FILE_WRITE:

    // read/write operations provide the offset in bytes.
    kernel_record->starting_file_page =
      file_offset >> PAGE_SIZE_ORDER;
    kernel_record->ending_file_page =
      (file_offset + length) >> PAGE_SIZE_ORDER;
    break;

  case TAG_MREMAP:
    {

    // mremap() provides a second virtual address range.
    kernel_record->old_starting_virtual_page =
      old_starting_virtual_address >> PAGE_SIZE_ORDER;
    kernel_record->old_ending_virtual_page =
      (old_starting_virtual_address + old_length) >> PAGE_SIZE_ORDER;

    // Sanity check: The length cannot overflow the address space.
    assert(kernel_record->old_starting_virtual_page <=
	   kernel_record->old_ending_virtual_page);    

    }
  }

  // For those record types that require the specification of a file
  // type, make the (unwarranted) assumption that a normal file was
  // used.  While inaccurate, it is what the v0.4.x series did.
  kernel_record->file_type = FILE_TYPE_REGULAR;

  kernel_record->record_number = kernel_record_number;

} // read_kernel_record
// ===================================================================
