/**
 * \file V0_5_Reader.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date December 2004
 * \brief Code for a class that reads kVMTrace v0.5 traces.
 *
 * A subclass of Reader that can read the reference and kernel traces
 * produced by kVMTrace v0.5.  See the header file for more details.
 */



// ===================================================================
// INCLUDES

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "V0_5_Reader.hh"
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
/**
 * \brief Whether the target architecture is big-endian.
 *
 * Since this reader handles binary files, it must account for the
 * endianness of the file data and its own internal data.
 * Specifically, big-endian targets will need to reverse the bytes of
 * each field read from the binary file.
 **/
const bool big_endian = (strcmp(HOSTTYPE, "powermac") == 0);
// ===================================================================



// ===================================================================
V0_5_Reader::V0_5_Reader (const char* const reference_trace_pathname,
			  const char* const kernel_trace_pathname) :
  Reader(reference_trace_pathname,
	 kernel_trace_pathname) {

  // Initialize the previous values (some fields of v0.5 reference
  // traces are difference encoded) to 0.
  previous_cycle_time = 0;
  previous_reference_time = 0;
  previous_context_ID = 0;

} // V0_5_Reader::V0_5_Reader
// ===================================================================



// ===================================================================
void
V0_5_Reader::read_reference_record
(reference_record_s* const reference_record) {

  reference_record_number++;

  // Attempt to read the next record of the reference trace.
  input_reference_record_s input_record;
  size_t read_result = fread(&input_record,
			     sizeof(input_record),
			     1,
			     reference_stream);

  // Did the read fail?
  if (read_result != 1) {

    // Distinguish between an EOF and other errors.
    if (feof(reference_stream)) {

      // And EOF simply indicates the end of the trace.  Artificially
      // set the tag accordingly.
      reference_record->tag = TAG_END_OF_TRACE;
      return;

    } else {

      // Other failures to read are the sign of a real problem.
      // Abort!
      fprintf(stderr, "V0_5_Reader::read_reference_record(): ");
      fprintf(stderr,
	      "Read error (code %d) at record %qx\n",
	      read_result,
	      reference_record_number);
      exit(1);

    }

  }

  // If this is a big-endian CPU, reverse the bytes of the record.
  input_reference_record_s* final_input_record = &input_record;
  if (big_endian) {
    final_input_record = reverse_input_bytes(&input_record);
  }

  // Copy the record type, timestamps, and virtual page number.
  reference_record->tag = final_input_record->tag;
  reference_record->cycle_timestamp =
    final_input_record->cycle_timestamp;
  reference_record->reference_timestamp =
    final_input_record->reference_timestamp;
  reference_record->virtual_page.context_ID =
    final_input_record->context_ID;
  reference_record->virtual_page.page_ID =
    final_input_record->virtual_page_number;

  // Store the record number.
  reference_record->record_number = reference_record_number;

} // V0_5_Reader::read_reference_record
// ===================================================================



// ===================================================================
void
V0_5_Reader::read_kernel_record (kernel_record_s* kernel_record) {

  // An ugly hack for when we want to skip one record and read the
  // next one.
 top_of_read_kernel_record:

  kernel_record_number++;

  // Attempt to read the next record of the kernel trace.
  char* read_result = fgets(buffer, BUFFER_SIZE, kernel_stream);

  // Did the read fail?
  if (read_result == NULL) {

    // If the EOF has been reached, then consider that a normal trace
    // ending.  Otherwise, abort processing.
    if (feof(kernel_stream)) {

      kernel_record->tag = TAG_END_OF_TRACE;
      kernel_record->record_number = kernel_record_number;
      return;

    } else {

      fprintf(stderr, "V0_5_Reader::read_kernel_record(): ");
      fprintf(stderr, "Read error kernel trace\n");
      exit(1);

    }
  }

  // Parse the record type, which should be the first character of the
  // buffer.
  kernel_record->tag = buffer[0];

  // Attempt to parse the rest of the record based on its type.
  // Declare and assign some variables here to avoid compiler
  // complaints of skipping initialization in switch-case statement.
  unsigned int parse_result = 0;
  unsigned int items_to_parse = 0;
  virtual_address_t starting_virtual_address = 0;
  virtual_address_t ending_virtual_address = 0;
  virtual_address_t old_starting_virtual_address = 0;
  offset_t length = 0;
  offset_t old_length = 0;
  file_offset_t file_offset = 0;
  switch (kernel_record->tag) {

  case TAG_SCHEDULE:
  case TAG_ACCEPT:

    items_to_parse = 3;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID);
    break;

  case TAG_EXIT:

    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->user_execution_time,
			  &kernel_record->system_execution_time);
    break;

  case TAG_FORK:

    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %hx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->parent_task_ID);
    break;

  case TAG_EXEC:

    items_to_parse = 7;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %hx %hx %s",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  kernel_record->filename);
    break;

  case TAG_CONTEXT_ASSIGNMENT:

    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->context_ID);
    break;

  case TAG_DUPLICATE_RANGE:

    items_to_parse = 6;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %lx %lx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->context_ID,
			  &kernel_record->duplicate_context_ID,
			  &starting_virtual_address,
			  &ending_virtual_address);
    break;

  case TAG_MMAP_FILE:

    items_to_parse = 11;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx %lx %hx %hx %qx %c %s",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  &file_offset,
			  &kernel_record->file_type,
			  kernel_record->filename);
    break;

  case TAG_MMAP_ANONYMOUS:

    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length);
    break;

  case TAG_MREMAP:

    items_to_parse = 7;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length,
			  &old_starting_virtual_address,
			  &old_length);
    break;

  case TAG_MUNMAP:

    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length);
    break;

  case TAG_COMPLETE_UNMAP:

    items_to_parse = 3;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->context_ID);
    break;

  case TAG_SHMAT:

    items_to_parse = 6;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length,
			  &kernel_record->shm_ID);
    break;

  case TAG_SHM_DESTROY:

    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->shm_ID);
    break;

  case TAG_COW_UNMAP:
  case TAG_DEBUG:
    // Note that debugging was used, in earlier v0.5 versions, to
    // differentiate between COW sites in the kernel.

    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address);
    break;

  case TAG_FILE_OPEN:

    items_to_parse = 9;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %hx %hx %qx %c %s",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  &kernel_record->file_size,
			  &kernel_record->file_type,
			  kernel_record->filename);
    break;

  case TAG_FILE_CLOSE:

    items_to_parse = 6;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %hx %hx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor);
    break;

  case TAG_FILE_READ:
  case TAG_FILE_WRITE:

    items_to_parse = 8;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %hx %lx %hx %hx %qx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  &file_offset,
			  &length);
    break;

  case TAG_FILE_DELETE:

    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %lx %hx %hx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor);
    break;

  case TAG_FILE_TRUNCATE:

    items_to_parse = 6;
    parse_result = sscanf(&buffer[2],
			  "%qx %qx %lx %hx %hx %qx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->reference_timestamp,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  &kernel_record->file_size);
    break;

  case TAG_END_OF_TRACE:

    fprintf(stderr, "V0_5_Reader::read_kernel_record(): ");
    fprintf(stderr, "Premature end of kernel trace\n");
    exit(1);

  case TAG_BLANK_LINE:

    // This may well be a blank line -- these occur sometimes due to a
    // goof in which file name logging inserts an excess newline.
    // Print a warning and repeat the reading of a record.
    fprintf(stderr,
	    "WARNING: Presumed blank line at record %qd -- skipping...\n",
	    kernel_record->record_number);
    goto top_of_read_kernel_record;

  default:
    fprintf(stderr, "V0_5_Reader::read_kernel_record(): ");
    fprintf(stderr,
	    "Invalid kernel tag %c(%d) at record %qd\n",
	    kernel_record->tag,
	    (int)kernel_record->tag,
	    kernel_record->record_number);
    exit(1);

  }

  // If the parse failed, emit the error and exit.
  if (parse_result != items_to_parse) {
    fprintf(stderr, "V0_5_Reader::read_kernel_record(): ");
    fprintf(stderr,
	    "Could not parse remainder at record %qd\n",
	    kernel_record_number);
    exit(1);
  }

  // Many record types specify either a virtual page or a range of
  // virtual pages.  Calculating the one/starting page can be done for
  // every record type, although it may be meaningless for some.  How
  // the ending page is calculated depends on record type, where some
  // provide a length of the range, and others provide an address for
  // the last page in the range.
  kernel_record->starting_virtual_page =
    starting_virtual_address >> PAGE_SIZE_ORDER;  

  // The length may not be page aligned, so align it by taking the
  // ceiling w.r.t. page size.
  offset_t page_aligned_length =
    (length + PAGE_SIZE - 1) & PAGE_NUMBER_MASK;
  offset_t old_page_aligned_length =
    (old_length + PAGE_SIZE - 1) & PAGE_NUMBER_MASK;
      
  switch (kernel_record->tag) {

  case TAG_DUPLICATE_RANGE:

    kernel_record->ending_virtual_page =
      ending_virtual_address >> PAGE_SIZE_ORDER;
    break;

  case TAG_MMAP_FILE:
  case TAG_MMAP_ANONYMOUS:
  case TAG_MREMAP:
  case TAG_SHMAT:
  case TAG_MUNMAP:

    // Convert the ending address to a page number.
    kernel_record->ending_virtual_page =
      ((starting_virtual_address + page_aligned_length)
       >> PAGE_SIZE_ORDER);

    // Sanity check: The length cannot overflow the address space.
    assert(kernel_record->starting_virtual_page <=
	   kernel_record->ending_virtual_page);

    break;

  }

  // For some record types, a second offset and length must be
  // converted into starting and ending pages.  For some, the address
  // range is within a file, and for others, it is a second range
  // within a virtual address space.
  switch (kernel_record->tag) {

  case TAG_MMAP_FILE:
    {

      // mmap() records provide the file offset in page numbers.
      unsigned int page_length = page_aligned_length >> PAGE_SIZE_ORDER;
      kernel_record->starting_file_page = file_offset;
      kernel_record->ending_file_page = file_offset + page_length;

    }
    break;

  case TAG_FILE_READ:
  case TAG_FILE_WRITE:

    // read/write operations provide the offset in bytes.
    kernel_record->starting_file_page =
      file_offset >> PAGE_SIZE_ORDER;
    kernel_record->ending_file_page =
      (file_offset + page_aligned_length) >> PAGE_SIZE_ORDER;
    break;

  case TAG_MREMAP:
    {

      // mremap() provides a second virtual address range.
      kernel_record->old_starting_virtual_page =
	old_starting_virtual_address >> PAGE_SIZE_ORDER;
      kernel_record->old_ending_virtual_page =
	((old_starting_virtual_address + old_page_aligned_length)
	 >> PAGE_SIZE_ORDER);

      // Sanity check: The length cannot overflow the address space.
      assert(kernel_record->old_starting_virtual_page <=
	     kernel_record->old_ending_virtual_page);    

    }
    break;

  }

  kernel_record->record_number = kernel_record_number;

} // V0_5_Reader::read_kernel_record
// ===================================================================



// ===================================================================
input_reference_record_s*
V0_5_Reader::reverse_input_bytes
(input_reference_record_s* source_record) {

  static input_reference_record_s reversed_record;

  reverse_object(&reversed_record.tag,
		 &source_record->tag,
		 sizeof(reversed_record.tag));
  reverse_object(&reversed_record.cycle_timestamp,
		 &source_record->cycle_timestamp,
		 sizeof(reversed_record.cycle_timestamp));
  reverse_object(&reversed_record.reference_timestamp,
		 &source_record->reference_timestamp,
		 sizeof(reversed_record.reference_timestamp));
  reverse_object(&reversed_record.context_ID,
		 &source_record->context_ID,
		 sizeof(reversed_record.context_ID));
  reverse_object(&reversed_record.virtual_page_number,
		 &source_record->virtual_page_number,
		 sizeof(reversed_record.virtual_page_number));

  return &reversed_record;

} // V0_5_Reader::reverse_input_bytes
// ===================================================================



// ===================================================================
void
V0_5_Reader::reverse_object (void* destination,
			     void* source,
			     unsigned int size) {

  // Take each buffer space as an array of characters.
  assert(sizeof(char) == 1);
  char* dst = (char*)destination;
  char* src = (char*)source;

  // Copy the bytes from the source to the destination in the reverse
  // of their original order.
  for (unsigned int i = 0; i < size; i++) {
    dst[size - i - 1] = src[i];
  }

} // V0_5_Reader::reverse_object
// ===================================================================
