/**
 * \file V0_5_Reader.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date December 2004
 * \brief Header of a class that reads kVMTrace v0.5 traces.
 *
 * A subclass of Reader that can read the reference and kernel traces
 * produced by kVMTrace v0.5.  The data from each record is then
 * transformed appropriately for use the merger utility.  See the
 * README file for more details on the reference trace format.
 */



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_V0_5_READER_HH)
#define _V0_5_READER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <stdio.h>
#include "Reader.hh"
// ===================================================================



// ===================================================================
// STRUCTURES
/**
 * \brief The source trace's record structure.
 *
 * The structure of the binary records written to a v0.5 source
 * reference trace.  We use this structure to read out each record
 * directly before converting it to our internal format.
 **/
struct input_reference_record_s {

  /** \brief The tag indicating the reference type. */
  tag_t tag;

  /** \brief The absolute cycle time of this reference. */
  timestamp_t cycle_timestamp;

  /** \brief The absolute reference time of this reference. */
  timestamp_t reference_timestamp;

  /** \brief The virtual address space in which the reference took
   *         place. */
  context_ID_t context_ID;

  /** \brief The virtual page number referenced. */
  virtual_page_ID_t virtual_page_number;

};
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * The V0_5_Reader definition.  This subclass of Reader that reads
 * kVMTrace v0.5 reference and kernel records.  It defines the
 * read_reference_record() and read_kernel_record() methods that
 * perform the actual reading, parsing, and processing of each record
 * before it is provided to the caller in a provided structure.
 **/
class V0_5_Reader : public Reader {

public: // CONSTRUCTOR

  /**
   * \brief The constructor.
   * \param reference_trace_pathname The pathname of the reference
   *        trace to be read.
   * \param kernel_trace_pathname The pathname of the kernel trace to
   *        be read. 
   * 
   * Set up a V0_5_Reader by initializing its cycle and reference
   * clocks and by opening the trace files and holding a streams for
   * each.
   **/
  V0_5_Reader (const char* const reference_trace_pathname,
	       const char* const kernel_trace_pathname);

public: // VIRTUAL METHODS

  /**
   * \brief Read the next record from the reference trace.
   * \see reference_record_s
   * \param reference_record A structure that will hold the needed
   *        data from the trace record.
   *
   * Read a single reference record, parse its data, process the
   * values, and place the needed results in the provided structure.
   * Any errors in reading (I/O errors or premature EOF) are expected
   * to abort processing.
   **/
  virtual void read_reference_record
  (reference_record_s* const reference_record);

  /**
   * \brief Read the next record from the kernel trace.
   * \see kernel_record_s
   * \param kernel_record A structure that will hold the needed
   *        data from the trace record.
   *
   * Read a single kernel record, parse its data, process the values,
   * and place the needed results in the provided structure.  Any
   * errors in reading (I/O errors or premature EOF) are expected to
   * abort processing.
   **/
  virtual void read_kernel_record
  (kernel_record_s* const kernel_record);

protected: // METHODS

  /**
   * \brief Reverse the bytes of an input reference record.
   * \see input_reference_record_s
   * \see big_endian
   * \param source_record The original reference record read directly
   *        from a binary file.
   * \return A new input reference record with the bytes of each field
   *         reversed from the source record.
   **/
  input_reference_record_s* reverse_input_bytes
  (input_reference_record_s* source_record);

  /**
   * \brief Reverse the bytes of a single object.
   * \param destination The buffer space into which the reversed bytes
   *        should be written.
   * \param source The buffer space of bytes to reverse.
   * \param size The number of bytes to reverse.
   **/
  void reverse_object (void* destination,
		       void* source,
		       unsigned int size);

protected: // DATA

  /** \brief The previous CPU cycle time. */
  timestamp_t previous_cycle_time;

  /** \brief The previous reference time. */
  timestamp_t previous_reference_time;

  /** \brief The previous context ID. */
  context_ID_t previous_context_ID;

}; // class V0_5_Reader
// ===================================================================



// ===================================================================
#endif // _V0_5_READER_HH
// ===================================================================
