// ===================================================================
/**
 * \file main.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date June 2004
 * \brief The entry point.
 *
 * This file is the entry point for the `merger' utility.  From the
 * command line arguments, it determines what type of merge is
 * requested, the format of the input traces, and then creates objects
 * for these choices that it then uses to perform the merge.
 **/
// ===================================================================



// ===================================================================
#include <string.h>
#include "types-and-constants.hh"
#include "Consumer.hh"
#include "Basic_Consumer.hh"
#include "Basic_Named_Consumer.hh"
#include "Per_Task_Consumer.hh"
#include "Per_Process_Consumer.hh"
#include "V0_4_Reader.hh"
#include "V0_5_Reader.hh"
// ===================================================================



// ===================================================================
/**
 * \brief The entry point.
 * 
 * \param argc The number of arguments given at the command line.
 * \param argv The vector of argument strings passed at the command
 *             line.
 * \return An execution result code.
 * 
 * Parse the command line arguments to obtain the type of merge
 * requested and the pathnames of the input traces.  Create and invoke
 * an object that will perform the requested type of merge.
 */

int
main (const unsigned int argc, const char** const argv) {

  // Were the correct number of arguments passed?
  if (argc != 6) {

    // No.  Emit usage and exit.
    fprintf(stderr, "Usage: %s\n", argv[0]);
    fprintf(stderr, "       <merge type {basic, basic-named, per-task, per-process}>\n");
    fprintf(stderr, "       <input trace type {v0.4, v0.5}>\n");
    fprintf(stderr, "       <reference trace pathname>\n");
    fprintf(stderr, "       <kernel trace pathname>\n");
    fprintf(stderr, "       <output pathname>\n");
    return 1;

  }

  initialize_constants();

  // Assign names to the arguments.
  const char* const merge_type = argv[1];
  const char* const input_trace_type = argv[2];
  const char* const reference_trace_pathname = argv[3];
  const char* const kernel_trace_pathname = argv[4];
  const char* const output_pathname = argv[5];

  // Determine what kind of trace is being used.
  Reader* reader = 0;
  if (strcmp(input_trace_type, "v0.4") == 0) {

    reader = new V0_4_Reader(reference_trace_pathname,
			     kernel_trace_pathname);

  } else if (strcmp(input_trace_type, "v0.5") == 0) {

    reader = new V0_5_Reader(reference_trace_pathname,
			     kernel_trace_pathname);

  } else {

    fprintf(stderr,
	    "main(): Unknown input trace type %s\n",
	    input_trace_type);
    exit(1);

  }

  // Ensure that the reference reader was allocated.
  if (reader == NULL) {

    fprintf(stderr, "main(): Unable to allocate reference reader\n");
    exit(1);

  }

  // Determine what kind of consumer (i.e. merge type, and thus output
  // type) was requested.
  Consumer* consumer = NULL;
  if (strcmp(merge_type, "basic") == 0) {

    consumer = new Basic_Consumer(reader, output_pathname);

  } else if (strcmp(merge_type, "basic-named") == 0) {

    consumer = new Basic_Named_Consumer(reader, output_pathname);

  } else if (strcmp(merge_type, "per-task") == 0) {

    consumer = new Per_Task_Consumer(reader, output_pathname);

  } else if (strcmp(merge_type, "per-process") == 0) {

    consumer = new Per_Process_Consumer(reader, output_pathname);

  } else {

    fprintf(stderr, "main(): Unknown merge type %s\n", merge_type);
    exit(1);

  }

  // Ensure that the consumer was allocated.
  if (consumer == NULL) {

    fprintf(stderr, "main(): Consumer allocation failed\n");
    exit(1);

  }

  // Merge the traces.
  consumer->merge();

  // Eliminate the consumer object.
  delete consumer;
  consumer = NULL;

  // Eliminate the reader object.
  delete reader;
  reader = NULL;

  // Exit successfully.
  return 0;

} // main
// ===================================================================
