/**
 * \file types-and-constants.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief The simple types and constants used in many modules and
 *        classes.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_TYPES_AND_CONSTANTS_HH)
#define _TYPES_AND_CONSTANTS_HH
// ===================================================================



// ===================================================================
// INCLUDES
#include <hash_map.h>
#include <list.h>
#include <stdio.h>
// ===================================================================



// ===================================================================
// FORWARD DECLARATIONS

struct Task;
struct Process;
struct Address_Space;
struct Canonical_Page;
// ===================================================================



// ===================================================================
// TYPES

/** \brief The type for the tag that indicates the type of a
 *         reference. */
typedef char tag_t;

/** \brief The type for a time counter. */
typedef unsigned long long int timestamp_t;

/** \brief The type for a virtual address. */
typedef unsigned long int virtual_address_t;

/** \brief The type for a reference length (bytes accessed). */
typedef unsigned short int length_t;

/**
 * \brief The type for a context (virtual address space) ID.
 *
 * This value is taken as the pointer to the top of the page table
 * that was active at the time of a reference.
 **/
typedef unsigned long int context_ID_t;

/**
 * \brief The type for a virtual page number.
 * \see PAGE_SIZE_ORDER
 *
 * This value is taken as some number of upper bits from a virtual
 * address.
 **/
typedef unsigned long int virtual_page_ID_t;

/** \brief The type for a canonical page number. */
typedef unsigned long int canonical_page_ID_t;

/** \brief The type for a task (thread) ID. */
typedef unsigned short int task_ID_t;

/** \brief The type for a process ID. */
typedef unsigned int process_ID_t;

/** \brief The type for an inode number, and thus a unique ID for a
 *         file. */
typedef unsigned long int inode_ID_t;

/** \brief The type for a device ID, uniquely identifying a device
 *         that holds files. */
typedef unsigned short int device_ID_t;

/** \brief The type of a IPC shared memory segment. */
typedef unsigned long int shm_ID_t;

/** \brief The type for an offset into a space, usually a virtual
 *         space. */
typedef unsigned long int offset_t;

/** \brief The type for an offset into a file. */
typedef unsigned long long int file_offset_t;

/** \brief The type for a file type. */
typedef char file_type_t;

/** \brief The type of the handler functions for the kernel events. */
class Consumer;
typedef void (Consumer::*handler_t) ();

/**
 * \brief The type for indicating the current merging phase.
 *
 * A merge passes through three distinct phases, and this data type
 * indicates which phase is the current one.  Specifically, a merge
 * will first be in a `pre-reference' phase, where only kernel events
 * are being processed, and no memory references have yet been
 * observed.
 *
 * Second, a merge will enter a `waiting-for-schedule' phase where a
 * reference has been observed, but no schedule has been processed
 * since.  Because schedulings are recorded only when references are
 * being recorded, there is a gap between the first references and the
 * first recorded scheduling during which the merger does not know to
 * which task to attribute the references.
 *
 * Finally, the merge passes into a `processing-references' phase
 * where references are being processed and correctly attributed to
 * their tasks.
 **/
typedef enum {
  PHASE_PRE_REFERENCE,
  PHASE_WAITING_FOR_SCHEDULE,
  PHASE_PROCESSING_REFERENCES
} merge_phase_t;
// ===================================================================



// ===================================================================
// CONSTANTS

/** \brief The size of buffer spaces allocated for temporary work. */
extern const unsigned int BUFFER_SIZE;

/**
 * \brief The number of offset bits in a virtual address.
 * \see PAGE_SIZE
 *
 * Given a page size n, the log_{2}(n), a.k.a. the number of lower
 * bits of an address to eliminate to create a page number.
 **/
extern const unsigned int PAGE_SIZE_ORDER;

/**
 * \brief The page size in bytes.
 * \see PAGE_SIZE_ORDER
 **/
extern const unsigned int PAGE_SIZE;

/**
 * \brief The ``canonical zero page''.
 *
 * The zero-page is a write-protected, permanently allocated page that
 * contains nothing for anonymously mapped pages that have not yet
 * been written. (Linux specific)
 **/
extern Canonical_Page canonical_zero_page;

/**
 * \brief The ``uncacheable page''.
 *
 * This page is not really a page at all, but a place-holder that can
 * be used to signify that an uncacheable page is being used (e.g. a
 * reference to a pipe), and thus should not necessarily be handled
 * normally.
 **/
extern Canonical_Page canonical_uncacheable_page;
 
/**
 * \brief A bit mask to eliminate the offset bits of an address.
 * \see PAGE_SIZE_ORDER
 **/
extern const unsigned int PAGE_NUMBER_MASK;

// Valid reference trace tags.
#define TAG_SINGLE_READ             'r'
#define TAG_MULTIPLE_READ           'R'
#define TAG_SINGLE_WRITE            'w'
#define TAG_MULTIPLE_WRITE          'W'

// Valid kernel trace tags.
#define TAG_SCHEDULE                'S'
#define TAG_FORK                    'F'
#define TAG_EXEC                    'X'
#define TAG_EXIT                    'T'
#define TAG_CONTEXT_ASSIGNMENT      'A'
#define TAG_DUPLICATE_RANGE         'D'
#define TAG_MMAP_FILE               'M'
#define TAG_MMAP_ANONYMOUS          'm'
#define TAG_MREMAP                  '@'
#define TAG_MUNMAP                  'U'
#define TAG_COMPLETE_UNMAP          'u'
#define TAG_COW_UNMAP               'C'
#define TAG_SHMAT                   's'
#define TAG_SHM_DESTROY             '/'
#define TAG_FILE_OPEN               '('
#define TAG_FILE_CLOSE              ')'
#define TAG_FILE_READ               '<'
#define TAG_FILE_WRITE              '>'
#define TAG_FILE_ELONGATING_READ    '{'
#define TAG_FILE_ELONGATING_WRITE   '}'
#define TAG_FILE_DELETE             '-'
#define TAG_FILE_TRUNCATE           't'
#define TAG_ACCEPT                  'a'
#define TAG_INVALID                 '!'
#define TAG_END_OF_TRACE            'e'
#define TAG_BLANK_LINE              (char)10
#define TAG_DEBUG                   '&'

// Valid file types.
#define FILE_TYPE_REGULAR           'f'
#define FILE_TYPE_PIPE              'p'
#define FILE_TYPE_DEVICE            'd'
#define FILE_TYPE_SOCKET            's'
#define FILE_TYPE_UNKNOWN           '!'

/**
 * \brief The kernel's context ID.
 *
 * This value is specific to Linux, which uses the address 0x101000 as
 * the top of the kernel's page table.
 **/
extern const context_ID_t KERNEL_CONTEXT;

/** \brief The kernel's process ID. */
extern const process_ID_t KERNEL_PROCESS_ID;

/** \brief The first page number of the kernel's virtual space that is
 *         shared with all tasks.  We use this value simply to verify
 *         that no recorded references are to kernel space. (Linux
 *         specific) */
extern const virtual_page_ID_t KERNEL_VIRTUAL_PAGE_OFFSET;

/** \brief The number of possible tag values. */
extern const int NUMBER_POSSIBLE_TAG_VALUES;
// ===================================================================



// ===================================================================
// STRUCTURES AND MAP TYPES

/** \brief A structure for uniquely identifying a virtual address
 *         within its virtual address space. */
struct virtual_page_s {

  context_ID_t context_ID;
  virtual_page_ID_t page_ID;

};

/** \brief A structure to hold a unique identifier for a file. */
struct file_ID_s {

  device_ID_t major; /**< The major device number. */
  device_ID_t minor; /**< The minor device number. */
  inode_ID_t inode_ID; /**< The inode number on the given device. */

  bool operator== (const file_ID_s& other) const {
    return ((major == other.major) &&
	    (minor == other.minor) &&
	    (inode_ID == other.inode_ID));
  }

  bool operator!= (const file_ID_s& other) const {
    return !operator==(other);
  }

};

/** \brief A structure for comparing task IDs used by tables indexed
 *         by task ID. */
struct equal_task_ID_s {
  bool operator() (const task_ID_t x, const task_ID_t y) {
    return (x == y);
  }
};

/** \brief A structure for comparing context IDs used by tables
 *         indexed by context. */
struct equal_context_ID_s {
  bool operator() (const context_ID_t x, const context_ID_t y) const {
    return (x == y);
  }
};

/** \brief A structure for comparing file IDs used by tables indexed
 *         by file ID. */
struct equal_file_ID_s {
  bool operator() (const file_ID_s& x, const file_ID_s& y) const {
    return (x == y);
  }
};

/** \brief A structure for comparing IPC shared memory segment IDs
 *         used by tables indexed by shm IDs. */
struct equal_shm_ID_s {
  bool operator() (const shm_ID_t x, const shm_ID_t y) const {
    return (x == y);
  }
};

/**
 * \brief A structure that provides a hashing function for file IDs.
 *
 * This function is required because file IDs are not scalar values
 * for which STL provides a default hashing operation.
 **/
struct file_ID_hash_s {

  size_t operator() (const file_ID_s& id) const {

    // Treat the value as a sequence of bytes, forming a hash from a
    // calculation borrowed from the strings hashing function in the
    // STL code itself.
    unsigned long int h = 0;
    char* s = (char*)&id;
    for (unsigned int i = 0; i < sizeof(file_ID_s); i++) {
      h = (5 * h) + *s;
      s++;
    }

    return (size_t)h;
  }

};

/** \brief A table type used to track the liveness of tasks, indexed
 *         by task ID. */
typedef hash_map<task_ID_t,
		 bool,
		 hash<task_ID_t>,
		 equal_task_ID_s> live_map_t;

/** \brief A table type used to map file IDs to address spaces. */
typedef hash_map<file_ID_s,
		 Address_Space*,
		 file_ID_hash_s,
		 equal_file_ID_s> file_ID_to_AS_map_t;

/** \brief A table type used to map IPC shared memory segment IDs to
 *         virtual->canonical mapping tables. */
typedef hash_map<shm_ID_t,
		 Address_Space*,
		 hash<shm_ID_t>,
		 equal_shm_ID_s> shm_ID_to_AS_map_t;

/** \brief A table type that maps task IDs to task objects. */
typedef hash_map<task_ID_t,
		 Task*,
		 hash<task_ID_t>,
		 equal_task_ID_s> task_ID_to_Task_map_t;

/** \brief A table type that maps contexts to task objects. */
typedef hash_map<context_ID_t,
		 Process*,
		 hash<context_ID_t>,
		 equal_context_ID_s> context_ID_to_Process_map_t;

/** \brief A structure to hold the contents of a reference trace
 *         record. */
struct reference_record_s {

  tag_t tag; /**< The type of reference. */
  timestamp_t cycle_timestamp; /**< The CPU cycle time of the
				  reference. */
  timestamp_t reference_timestamp; /**< The reference count at the
				      time of the reference. */
  virtual_page_s virtual_page; /**< The virtual page referenced. */
  unsigned long long int record_number; /**< The record number within
					   the reference trace. */

};

/**
 * \brief A structure to hold the contents of a kernel trace record.
 *
 * Depending on the type of event represented, some fields may not be
 * used or may have slightly different uses.
 **/
struct kernel_record_s {

  kernel_record_s () {

    // Allocate filename space.
    filename = new char[BUFFER_SIZE];
    if (filename == NULL) {
      fprintf(stderr, "kernel_record_s::kernel_record_s(): ");
      fprintf(stderr, "Could not allocate filename space\n");
      exit(1);
    }

  }

  tag_t tag; /**< The type of event. */
  timestamp_t cycle_timestamp; /**< The global CPU cycle time of the
				  event. */
  timestamp_t reference_timestamp; /**< The virtual reference cycle
				      time of the event for the
				      task. */
  task_ID_t task_ID;  /**< The task that caused or was the primary
			 target of the event. */
  task_ID_t parent_task_ID; /**< The task that requested a fork(),
			       where the task_ID field is the newly
			       forked process. */
  context_ID_t context_ID; /**< The context associated with the
			      event. */
  context_ID_t duplicate_context_ID; /**< For a context duplication,
					the destination context, where
					the previous field is the
					source context. */
  virtual_page_ID_t starting_virtual_page; /**< The first (and
				              sometimes only) virtual
				              page for the event. */
  virtual_page_ID_t ending_virtual_page; /**< The last virtual page
				            for those events that
				            specify a virtual
				            range. */
  virtual_page_ID_t old_starting_virtual_page; /**< The first virtual
						  page for an old
						  mapping (see
						  mremap). */
  virtual_page_ID_t old_ending_virtual_page; /**< The last virtual page
						for an old mapping. */
  file_ID_s file_ID; /**< The unique file ID associated with the
			event. */
  virtual_page_ID_t starting_file_page; /**< The first (and sometimes
					   only) page within the
					   file's address space. */
  virtual_page_ID_t ending_file_page; /**< The last page within the
					 file for events that specify
					 a range. */
  file_offset_t file_size; /**< The length of a file, given in
				  pages. */
  shm_ID_t shm_ID; /**< The IPC shared memory segment ID associated
		      with the event. */

  file_type_t file_type; /**< The type of a file. */

  unsigned long int user_execution_time; /**< The total number of CPU
					    seconds spent in
					    user-level code. */

  unsigned long int system_execution_time; /**< The total number of
					      CUP seconds spent in
					      system-level code. */

  char* filename; /**< The name of the file for this operation.  Note
		        that this identifier is not unique as the file
		        ID is, but indicates the name by which the
		        process accessed the file. */

  unsigned long long int record_number; /**< The record number within
					   the reference trace. */

};
// ===================================================================



// ===================================================================
// FUNCTIONS

/**
 * \brief Initialize not-fully-configured constants.
 *
 * Some constants may require more than an initial assignment to be
 * completely initialized.  Complete their initial configuration.
 **/
void initialize_constants ();
// ===================================================================



// ===================================================================
#endif // _TYPES_AND_CONSTANTS_HH
// ===================================================================
